;;; ************************************************************************
;;; Emacs Lisp code for C++
;;; Author: Wade Holst, 1995
;;; CopyLeft
;;; ************************************************************************
;;; 
;;; Some notes:
;;;   - If you use this code, there is one fundamental key-binding modification
;;;     that might be problematic.  In emacs, one of the crucial operations is
;;;     that of "setting the mark", which is bound to Ctrl-Space (which is denoted
;;;     C-@ in emacs).  Setting the mark is just emacs-speak for remembering
;;;     the current cursor position.  There are hundreds of emacs commands that
;;;     use a previously set mark in order to operate properly, so the set-mark
;;;     command (keystroke C-@) is rather important.
;;; 
;;;   - However, if you use this c++.el file, C-@ will no longer set the mark. 
;;;     Instead, to set the mark you will have to press C-@ twice
;;;     (i.e. C-@ C-@).  This is because C-@ has been made its own
;;;     keymap, and all of the key-bindings for doing C++ discussed
;;;     below start with C-@.
;;; 
;;;   - I've only tested this code with xemacs - it may not (yet) work 
;;;     with emacs.  Remember that on the gaul machines, you should
;;;     invoke xemacs using /usr/local/bin/xemacs (the default version
;;;     is old and sloooowwwww to load)
;;; 
;;;   - There is various additional functionality that could be added.  If
;;;     you would like to see something in particular, let me know.
;;; 
;;; 
;;; Here are the key-bindings provided by the emacs code:    
;;; 
;;;   SHELLS
;;;   ======
;;; 
;;;   These key-bindings allow you to create and easily access subshells within
;;;   emacs:
;;; 
;;;     C-@ s 1
;;;     C-@ s 2
;;;     C-@ s 3
;;;     C-@ s 4
;;;     C-@ s 5
;;; 
;;;   Subshells in emacs are extremely useful because they allow you to navigate
;;;   in your directory hierarchy while staying with the consistent environment
;;;   provided by emacs.  However, there are some subtle issues that you should
;;;   be aware of.  Most importantly, subshells in emacs are only useful if
;;;   emacs does a good job of keeping track of the "current" directory.  If
;;;   you have a subshell inside emacs and execute the following:
;;;      obelix% cd ~/Some/Directory
;;;   there are TWO different "current directories"
;;;      1) the current directory within the subshell
;;;      2) the current directory that emacs uses (for example, when providing a
;;;         default path when you try to load a new file with C-x C-f).
;;; 
;;;   Emacs tries to keep its idea of the current directory synchronized
;;;   with that of the subshell the cursor is currently in, but it
;;;   sometimes gets confused.  You can always resynch the two versions
;;;   with the command:
;;;      C-@ .
;;;   which should be invoked while the cursor is in a window representing
;;;   a subshell.
;;; 
;;; 
;;;   CREATING FILES 
;;;   ==============
;;; 
;;;   These key-bindings create files (for example, main.cc, Makefile,
;;;   X.cc, X.hh, etc.)
;;;   
;;;     Main program  - creates a stub main.cc
;;;       C-@ c f m
;;;   
;;;     Makefile      - creates a generic Makefile
;;;       C-@ c f M
;;;   
;;;     Class  (class name will be prompted for in minibuffer - creates Name.hh and Name.cc)
;;;       C-@ c f c
;;;   
;;;   
;;;   ACCESSORS
;;;   =========
;;; 
;;;   These key-bindings automate the process of defining state and the
;;;   associated accessors.  They rely on the .hh and .cc files having the
;;;   format created by C-@ c f c
;;; 
;;;     Creating state whose accessors should pass by reference (the most common case)
;;;        C-@ c i v
;;; 
;;;     Creating state whose accessors should pass by value (common for primitive types)
;;;        C-@ c i V
;;; 
;;;     Creating state whose code is in the .cc file
;;;        C-@ c I v
;;;        C-@ c I V
;;; 
;;;   WINDOWS
;;;   =========
;;; 
;;;   These bindings automate the process of loading up one or more windows with
;;;   particular files.
;;; 
;;;      Viewing hh/cc vertically (.hh on top, .cc on bottom)
;;;        C-@ c w v
;;;   
;;;      Viewing hh/cc horizontally (.hh on left, .cc on right)
;;;        C-@ c w h
;;; 
;;;      Only windows are hh/cc/shell (vertically)
;;;        C-@ c W v
;;; 
;;;      Only windows are hh/cc/shell (horizontally)
;;;        C-@ c W h
;;;   
;;;   MISCELLANEOUS
;;;   =============
;;; 
;;;     You can turn font highlighting on/off with the command
;;;        C-@ h
;;;     This is useful in situations where font highlighting isn't turned on
;;;     automatically and when emacs hasn't properly refreshed its highlighting 
;;;     after esoteric code modifications.  
;;; 
;;; 
;;;     You can insert the ifdef/endif and class stub into a file with
;;;        C-@ c i h
;;;     However, this is usually completely unnecessary, since C-@ c f c
;;;     does that and much more.
;;; 
;;; 
;;; And here is the lisp code.  Rremember to put it in ~/.c++.el and add 
;;; the following to your ~/.emacs file:
;;;    (load-file "~/.c++.el")
;;; 
;;; ************************************************************************

(if (not (boundp 'my-map))
  (let ()
    (setq my-map (make-sparse-keymap))
    (define-key global-map "" 'eval-expression)

    (define-key global-map [?\C-\ ] my-map)       
    (define-key global-map "\C-@" my-map)         

    (define-key my-map [?\C-\ ] 'set-mark-command)
    (define-key my-map "\C-@" 'set-mark-command)
  )
)

(if (not (boundp 'running-xemacs)) (setq running-xemacs nil))

(defun highlight ()
   "Toggle highlighting on/off in current buffer"
   (interactive)
   (if font-lock-mode
       (font-lock-mode 0)
       (progn (font-lock-mode 1) (font-lock-fontify-buffer))
   )
)
(define-key my-map "h" 'highlight)


(defun my-find-file (filename)
  "Edit file FILENAME.
Switch to a buffer visiting file FILENAME,
creating one if none already exists.

Reloads file from disk. "

  (interactive "FFind file: ")
  (let ((buf (get-file-buffer filename)))
    (if (null buf)
      (find-file filename) 
	  ; else
      (let ()
		(switch-to-buffer buf)
      	(if (buffer-modified-p buf)
      	  (find-file filename)  
      	  ; else
      	  (revert-buffer t t)
        )
      )
    )
  )
)   


;;; ************************************************************************
;;; Lisp code for Shell Editing
;;; ************************************************************************

(defun edit-shell-general (shell)
   "Edit the shell with given name 'shell'.  If such a shell buffer
    doesn't exist, create it and add it to the global variable
    G-shell-list."
   (let ()
      (if (not (null (get-buffer shell)))
         (switch-to-buffer shell)
       ; else
         (progn
            (shell)
	    ; (setq tab-width 8)
            (rename-buffer shell)
         )
      )
   )
)

(define-key my-map "s1" '(lambda () (interactive) (edit-shell-general "Shell-1")))
(define-key my-map "s2" '(lambda () (interactive) (edit-shell-general "Shell-2")))
(define-key my-map "s3" '(lambda () (interactive) (edit-shell-general "Shell-3")))
(define-key my-map "s4" '(lambda () (interactive) (edit-shell-general "Shell-4")))
(define-key my-map "s5" '(lambda () (interactive) (edit-shell-general "Shell-5")))

(if running-xemacs
  (let()
    (define-key my-map "." 'shell-resync-dirs)
  )
)

;;; ************************************************************************
;;; Lisp code for C and C++ Editing
;;; ************************************************************************

(defvar c++-current-module "Misc")

;; Each of these variables encodes three pieces of information:
;;     - the suffix of a file to access
;;     - a regexp to search for.  If empty (i.e. ""), it means move to the
;;       end of the file.
;;     - a number of lines to move, from the end of the regexp.  To move,
;;       up, use negative numbers

;; Assumptions about file format:
;;   - class definitions have at least one 'public:', 'protected:' and 'private:'
;;     section (explicitly stated)

(defvar c++-header-suffix ".h")
(defvar c++-pheader-suffix ".h") ; sometimes ".pih" instead
(defvar c++-source-suffix ".cc")  ; sometimes ".cpp" or ".C" or ".CC"
(defvar c++-inline-suffix ".cc")  ; sometimes ".icc" instead

(defvar c++-public-header-spec     (list c++-header-suffix "public:" 1))
(defvar c++-protected-header-spec  (list c++-pheader-suffix "protected:" 1))
(defvar c++-private-header-spec    (list c++-pheader-suffix "private:" 1))

(defvar c++-inline-source-spec (list c++-inline-suffix "ACCESSORS" 1))

(defun c++-current-file-base ()
   "Returns the base name (i.e. without suffix) of the file associated
    with the current buffer"
   (wmh-replace-in-string (buffer-file-name) "\\.[^/]*$" "")
)  

(defun c++-current-class ()
   "Returns the name of the class (based on the file)"
   (let ( (file (buffer-file-name)))
      (if (string-match ".*/\\([^\.]+\\)" file)
         (match-string 1 file)
       ; else
         nil
      )
   )
)

(defun c++-make-class (name) 
   (interactive "sName: ")
   (let ((hh (concat name ".h"))
         (cc (concat name ".cc"))
         (icc (concat name ".icc"))
        )

   (find-file hh)
   (insert (format "
# ifndef %s_hh
# define %s_hh %s_hh

class %s {
public:
    // Constructors/Destructors
    %s();

    // Accessors

    // Input/Output

    // Miscellaneous

protected:
    // Accessors

    // Methods 
 
private: 
    // Accessors

    // Methods 

    // State
};

# endif // %s_hh

" name name name name name name) 
)
    (save-buffer)
    (split-window-vertically)
    (other-window 1)
    (find-file cc)
    (insert (format "
# include <iostream>
# include \"%s.h\"  

// ******************************************************************
// CONSTRUCTORS

%s::%s() {
    // Default Constructor
}

// ******************************************************************
// ACCESSORS


// ******************************************************************
// METHODS

"
    name name name))
    (save-buffer)
  )
)
(define-key my-map "cfc" 'c++-make-class)

(defun c++-load-hh (base)
  "Load base.h and highlight it"
  (my-find-file (concat base ".h"))
  (highlight)
  (highlight)
)

(defun c++-load-cc (base)
  "Load base.cc and highlight it"
  (my-find-file (concat base ".cc"))
  (highlight)
  (highlight)
)

(defun c++-load-hh-cc-tb (base)
   "Split current window into two windows containing 
    the .cc and .h files for the given base name"
   (interactive "sBase: ")
   (split-window-vertically)
   (c++-load-hh base)
   (other-window 1)
   (c++-load-cc base)
)
(define-key my-map "cwv" 'c++-load-hh-cc-tb)

(defun c++-load-hh-cc-lr (base)
   "Split current window into two windows containing 
    the .cc and .h files for the given base name"
   (interactive "sBase: ")
   (split-window-horizontally)
   (c++-load-hh base)
   (other-window 1)
   (c++-load-cc base)
)
(define-key my-map "cwh" 'c++-load-hh-cc-lr)

(defun c++-load-hh-cc-sideways (base)
   "Create three windows containing the .cc and .h files for the given base name"
   (interactive "sBase: ")
   (delete-other-windows)
   (split-window-horizontally)
   (c++-load-hh-cc-tb base)
   (other-window 1)
   (edit-shell-general "Shell-1")
)
(define-key my-map "cWv" 'c++-load-hh-cc-sideways)

(defun c++-load-hh-cc-stacked (base)
   "Create three windows containing the .cc and .h files for the given base name"
   (interactive "sBase: ")
   (delete-other-windows)
   (split-window-vertically)
   (c++-load-hh-cc-lr base)
   (other-window 1)
   (edit-shell-general "Shell-1")
)
(define-key my-map "cWh" 'c++-load-hh-cc-stacked)

(defun c++-goto-spec (spec &optional sub)

   "Given a file/position specification (see documentation at
    beginning of c++.el), place the cursor at the beginning of the
    appropriate line in the specified file.  If optional SUB is
    provided, it is a (partial) regexp indicating a line to search for
    from the line indicated by spec.  The regexp \"//\[ \t\]*\" is
    prepended to SUB to form the actual regexp.  If the searched-for
    line does not exist, point stays at the position dictated by spec.

    Returns the current buffer at time of invocation"

   (let ((buf (current-buffer))
         (suffix (car spec))
         (regexp (car (cdr spec)))
         (lines (car (cdr (cdr spec))))
         (base (c++-current-file-base))
		 (res t)
        )
        
     ;; Obtain correct file
     (find-file (concat base suffix))

     ;; Move to beginning of file
     (goto-char (point-min))

     ;; If regexp provided, search for it
     (if (equal regexp "")
	 (goto-char (point-max))
       ; else
         (setq res (re-search-forward regexp (point-max) t))
      )

     ;; Move forward specified number of lines
     (if res (next-line lines))
     (beginning-of-line)

     ;; if subregexp provided, search for it
     (if (not (null sub)) 
        (let () 
           (re-search-forward (concat "//\[ \t\]*" sub) (point-max) t)	 
           (next-line 1)
           (beginning-of-line)
        )
     )

     ; Return the buffer
     buf
   )
)

(defun c++-goto-variable-spec (assc key default &optional sub)
   "Goto a position in a file based on a position specification
    established by the value of KEY in ASSC.  For example, if
    ASSC is ((get 2) (set 1) (setref 1) (var 0)) and KEY is 'set,
    the value will be 1, which corresponds to c++-protected-header-spec.
 
    0 is c++-private-header-spec
    1 is c++-protected-header-spec
    2 is c++-public-header-spec

   -1 means don't do anything

   Returns t if successful, nil if failure (i.e. when -1 encountered)
   "

   (let ( (value (cadr (assoc key assc)))
          spec
        )
       (if (null value) (setq value default))

       (setq spec
          (if (eq value 2)
	      c++-public-header-spec
              (if (eq value 1)
		 c++-protected-header-spec
                 (if (eq value 0)          
		     c++-private-header-spec
		     nil
                 ))))

       (if (not (null spec))
          (c++-goto-spec spec sub)
       )

       ;; Return value
       (not (null spec))
   )
)

(defun c++-insert-state (type var &optional scopeassoc)
   "Given a variable name, automatically generate public and private
    declarations in the .h file, and associated definitions in the
    .cc/.icc file.

    Optional SCOPEASSOC provides indications of how certain variations
    should be handled.  Keys are:
       ref     --> 0 means do not pass variable around by reference, 1 means pass by reference (default is 1)
       get     --> 2 means public, 1 means protected, 0 means private, -1 means don't generate
       set     --> 2 means public, 1 means protected, 0 means private, -1 means don't generate
       setref  --> 2 means public, 1 means protected, 0 means private, -1 means don't generate 
       var     --> 2 means public, 1 means protected, 0 means private, -1 means don't generate 
       inline  --> a string (empty string by default).  Usually 'inline ' or 'INLINE ' if specified.
    "

   (interactive "sType: \nsVariable: ")
   (let* (buf           (current-buffer)                             ; buffer
         class                                                       ; string 
         (refinfo       (cadr (assoc 'ref scopeassoc)))              ; integer or nil
         (ptr-p         (and (null (string-match "<" type)) (not (null (string-match "*" type)))))  ; boolean (this doesn't work fully, because we abandon attempts for templates.
         (ref           (and (not ptr-p) (or (null refinfo) (eq refinfo 1))))   ; boolean
         (capvar        (my-capitalize var))                            ; string
         (refstr        " &") ; (if ptr-p "" " &"))                          ; string
         (reftypedecl   (c++-module-type c++-current-module (if ref (concat type refstr) type)))  ; string
         (refretdecl    (c++-module-type c++-current-module (concat type refstr)))                ; string
         (consttypedecl (c++-module-type c++-current-module (concat "const " (if ref (concat type refstr) type))))  ; string
	 (voiddecl      (c++-module-type c++-current-module "void")) ; string
         (inline        (if (null (assoc 'inline scopeassoc)) "inline " (cadr (assoc 'inline scopeassoc)))) ; string
         get-p    ; boolean
         set-p    ; boolean
         setref-p ; boolean
         var-p    ; boolean
         (inline-p      (if (> (length inline) 0) t nil))
	 (get-body (format " { return this->_%s; }" var))
         (set-body (format " { this->_%s = %s; }" var var))
         (setref-body (format " { return this->_%s; }" var))
        )

     (if (not inline-p)
        (setq inline "inline ")
     )

     ;; Get accessor
     (if (c++-goto-variable-spec scopeassoc 'get 2 "Accessors")
       (let ()
        (insert (format "    %s%-15s %s() const%s\n" inline consttypedecl var (if inline-p get-body ";")))
        (setq get-p t) 
       )
     )

     ;; Set accessor
     (if (c++-goto-variable-spec scopeassoc 'set 1 "Accessors")
       (let ()
	 (insert (format "    %s%-15s %sIs(%s %s)%s\n" inline voiddecl var (if ptr-p type consttypedecl) var (if inline-p set-body ";")))
         (setq set-p t)
       )
     )

     ;; Setref accessor
     (if (c++-goto-variable-spec scopeassoc 'setref 0 "Accessors")
       (let ()
	 (insert (format "    %s%-15s %sRef()%s\n" inline refretdecl var (if inline-p get-body ";")))
         (setq setref-p t)
       )
     )

     ;; Variable declaration
     (if (c++-goto-variable-spec scopeassoc 'var 0 "State")
       (let ()
	 (insert (format "    %-15s _%s;\n" type var))
         (setq var-p t)
       )
     )

     ;; Obtain the name of the class!
     (re-search-backward "\n[ \t]*class")
     (forward-word 2) ;; skip past 'class' and name
     (setq class (current-word)) ;; the name of the class, just before point

     (if (not inline-p)
       (let ()
         ;; Add code corresponding to the above public and private methods
     	 (c++-goto-spec c++-inline-source-spec)
	 (insert "\n")
     	 (if get-p 
     	    (insert (format "%s %s::get%s() const%s\n" 
   			 consttypedecl class capvar get-body ))
     	 )
    
     	 (if set-p 
     	    (insert (format "%s %s::set%s(%s %s)%s\n" 
  				voiddecl class capvar (if ptr-p type consttypedecl) var set-body ))
     	 )
    
     	 (if setref-p
     	    (insert (format "%s %s::%s()%s\n" 
   				refretdecl class var setref-body ))
     	 )   

        ;; Switch back to the original buffer
        (switch-to-buffer buf)
      )
     )
   )
)
(define-key my-map "civ" 'c++-insert-state)

(defun c++-insert-state-special (type var &optional scopeassoc)
   "As c++-insert-state but inline is set automatically to ''"
   (interactive "sType: \nsVariable: ")  
   (if (null (assoc 'inline scopeassoc))
       (setq scopeassoc (append scopeassoc '((inline ""))))
   )
   (c++-insert-state type var scopeassoc)
)
(define-key my-map "cIv" 'c++-insert-state-special)

(defun c++-insert-primitive-state (type var &optional scopeassoc)
  "Provides some defaults for scopeassoc"
   (interactive "sType: \nsVariable: ")
  (if (null (assoc 'ref scopeassoc))
      (setq scopeassoc (append scopeassoc '((ref 0))))
  )
  (c++-insert-state type var scopeassoc)
)
(define-key my-map "ciV" 'c++-insert-primitive-state)

(defun c++-insert-primitive-state-special (type var &optional scopeassoc)
   "As c++-insert-primitive-state but inline is set automatically to ''"
   (interactive "sType: \nsVariable: ")  
   (if (null (assoc 'inline scopeassoc))
       (setq scopeassoc (append scopeassoc '((inline ""))))
   )
   (c++-insert-primitive-state type var scopeassoc)
   
)
(define-key my-map "cIV" 'c++-insert-primitive-state-special)

(defun c++-module-type (module type)
   "Return a string representing a module-specific type for use in
    accessor definitions.  The module component is usually of the
    form  <module>_INLINE, and is a preprocessor constant that is
    either 'inline' or '' depending on whether inlining is desired
    or not.  In simple cases, this method will just return the type
    without adding any module info"

   ; (concat module "_INLINE " type)
    type
)


(defun c++-file-main ()
   (interactive)
   (let ()
      (find-file "main.cc")
      (insert (format "
using namespace std;
# include <iostream>

int
main(int argc, char **argv) {

    cout << \"Hello, World\" << endl;
    return(0);
}
"))
      (next-line -2)
      (c-indent-command)
      (save-buffer (current-buffer))
   )
)
(define-key my-map "cfm" 'c++-file-main)

(defun c++-module-set (module)
   (interactive "sModule: ")
   (setq c++-current-module module)
)
(define-key my-map "cms" 'c++-module-set)

(defun c++-insert-header() 
   (interactive)
   (let ((file (wmh-replace-in-string (buffer-file-name) 
		  "\\(.*/\\)\\([^/]+\\)\\.\\(h+\\)$"
		  "\\2")))
      (insert (format "# ifndef %s_h\n# define %s_h %s_h\n\nclass %s {\n\n};\n\n# endif // %s_h\n"
					  file file file file file))
	  (next-line -4)
   )
)
(define-key my-map "cih" 'c++-insert-header)

(defun c++-file-makefile ()
  (interactive)
  (find-file "Makefile")
  (if (re-search-forward "# ORG:" (point-max) t)
	  (let () (beginning-of-line) (kill-line 1)))
  (goto-char (point-max))

  (insert "
############################################################################
#                              Use gmake!
############################################################################

# Change this if you want the executable named something else.
EXEC       = main

# Add compilation and linking flags (and include/libary paths) here
DEBUG      = -g
CFLAGS     = $(DEBUG) -Wall
LDFLAGS    = $(DEBUG)
LIBS       = 
INCS       =  

# Change these if src code is in a subdirectory or if the 
# default installation directory is not desired.
ROOTDIR    = .
SRCDIR     =  $(ROOTDIR)
INSTALLDIR =  $(HOME)/root

# You should not need to change these variables (but can if you feel
# up to it).
SRCS       = $(shell find $(SRCDIR) -name '*.cc')
OBJS       = $(SRCS:.cc=.o)
INCDIRS   := $(shell find $(SRCDIR) -name 'CVS' -prune -o -type d -print)
CFLAGS    += $(addprefix -I,$(INCDIRS))
LDFLAGS   += $(LIBS)

CC         = g++

# The generic .cc to .o compilation.  Note that the 'dep' file
# contains the dependencies for each .o file, so that files are always
# recreated when necessary.
.cc.o:
	$(CC) -o $@ -c $(CFLAGS) $<

# The executable rule
$(EXEC): $(OBJS) $(INCS) dep
	$(CC) -o $(EXEC) $(LDFLAGS) $(OBJS)

# Install the executable
install: $(EXEC)
	install -d $(INSTALLDIR)
	install --mode=755 $(EXEC) $(INSTALLDIR)

# All intermediate files sould be removed here.  By default, this
# consists of all object files.  Add any other intermediate files you
# deem appropriate.
clean:
	rm -f $(OBJS)

# All files that are created by this makefile are removed here.
# By default, that is all files removed by 'clean' plus the executable
# and 'dep' files. Add any other generated files.
clobber: clean
	rm -f $(EXEC) dep

# Forced dependency creation.  The 'dep' file contains .o targets,
# with all files that the .o target depends on automatically computed
# for us.  This allows us to ensure that a particular object file is
# regenerated whenever any file it depends on is modified.
depend: FORCE
	$(CC) -MM $(SRCS) | perl -pe 's%^(\\S+)\\.o:\\s+(\\S+)/([^/]+).cc%$$2/$$1.o: $$2/$$3.cc%;' > dep

# Non-forced dependency creation
dep:
	$(CC) -MM $(SRCS) | perl -pe 's%^(\\S+)\\.o:\\s+(\\S+)/([^/]+).cc%$$2/$$1.o: $$2/$$3.cc%;' > dep

FORCE:

# This includes the 'dep' file so that .o dependencies are considered
# by the makefile.
-include dep

")
  (save-buffer (current-buffer))
)
(define-key my-map "cfM" 'c++-file-makefile)


(defun c++-demangle () 
   "Demangle the mangled function name at point.  
Assumes that point is in a sub-shell."
   (interactive)

   (re-search-backward "\[\t\n \]")
   (let ((s (+ (point) 1)) str str2 (buf (current-buffer)))
      (forward-char 1) 
      (re-search-forward "\[\t\n \]")
      (forward-char -1)
      (setq str (buffer-substring s (point)))
      (shell-command (concat "g++filt " str))
      (switch-to-buffer "*Shell Command Output*")
      (replace-regexp "\n" "")
      (setq str2 (buffer-substring (point-min) (point-max)))
      (kill-buffer (current-buffer))
      (switch-to-buffer buf)
      (message str2)
   )
)
(define-key my-map "cmm" 'c++-demangle)

(defun c++-find-file-around-point (prefix)
   (interactive "p")
   "Given a file name at point, load that file in the 'previous'
    window (with respect to the window in which this command was invoked)."

   (if (not (= prefix 4)) 
      (re-search-backward "\[\t\n \]")
      (forward-char -1)
   )

   (forward-char 1) 
   ; Ignore ../ (Cxx stuff)
   (setq A (buffer-substring (point) (+ 3 (point))))
   (if (equal (buffer-substring (point) (+ 3 (point))) "../")
      (forward-char 3)
   )

   (let ((s (point)) str (buf (current-buffer)))
      (re-search-forward "\[:\t\n \]")
      (forward-char -1)
      (setq str (buffer-substring s (point)))
      (setq Str str)
      (other-window-backward)
      (switch-to-buffer buf)  ; so that we have the same relative path
      (message str)
      (find-file str)
      (other-window 1)
      (goto-char s)
      (other-window -1)
   )
)
(define-key my-map "cff" 'c++-find-file-around-point)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;(require 'c-mode)
(setq-default
   c-echo-syntactic-information-p  t
   c-basic-offset                  2
   cperl-indent-level              2 
)

(setq c-basic-offset 2)

(setq-default      c-offsets-alist
  '((string                . -1000)
    (c                     . c-lineup-C-comments)
    (defun-open            . 0)
    (defun-close           . 0)
    (defun-block-intro     . +)
    (class-open            . 0)
    (class-close           . 0)
    (inline-open           . 0)
    (inline-close          . 0)
    (ansi-funcdecl-cont    . +)
    (knr-argdecl-intro     . +)
    (knr-argdecl           . 0)
    (topmost-intro         . 0)
    (topmost-intro-cont    . 0)
    (member-init-intro     . +)
    (member-init-cont      . 0)
    (inher-intro           . +)
    (inher-cont            . c-lineup-multi-inher)
    (block-open            . 0)
    (block-close           . 0)
    (brace-list-open       . 0)
    (brace-list-close      . 0)
    (brace-list-intro      . +)
    (brace-list-entry      . 0)
;; choose one of these
    (statement             . 0)
   ;(statement             . c-lineup-runin-statements)
    (statement-block-intro . +)
    (substatement          . +)
    (substatement-open     . 0)
;; choose one of these
    (statement-cont        . +)
   ;(statement-cont        . c-lineup-math)
    (case-label            . 2)
    (statement-case-intro  . 2)
    (statement-case-open   . 0)
    (access-label          . -)
    (label                 . 2)
    (do-while-closure      . 0)
    (else-clause           . 0)
    (comment-intro         . c-lineup-comment)
    (arglist-intro         . +)
    (arglist-cont          . 0)
    (arglist-cont-nonempty . c-lineup-arglist)
    (arglist-close         . +)
    (stream-op             . c-lineup-streamop)
    (inclass               . +)
    (cpp-macro             . -1000)
    (friend                . 0)
    (objc-method-intro     . -1000)
    (objc-method-args-cont . c-lineup-ObjC-method-args)
    (objc-method-call-cont . c-lineup-ObjC-method-call)
))


(setq c++-header-suffix ".h")
(setq c++-source-suffix ".cc")
(setq c++-inline-suffix ".icc")

(setq c++-decl-match "^\\([ \t]*\\)\\(inline[ \t]*\\)?\\(.*[^ \t]\\)\\([ \t]+\\)\\([a-zA-Z0-9_]+\\)\\([ \t]*\\)\\((.*)\\)\\([ \t]*\\)\\(const\\)?")
(setq c++-decl-vars  '(indent opt-inline type ws-type-name name ws-name-args args ws-args-post post))

(setq c++-defn-match "^\\([ \t]*\\)\\(.*[^ \t]\\)\\([ \t]+\\)\\([a-zA-Z0-9_]+[ \t]*::[ \t]*\\)\\([a-zA-Z0-9_]+\\)\\([ \t]*\\)\\((.*)\\)\\([ \t]*\\)\\(\\|const\\)?[ \t]*\{?[ \t]*$")
(setq c++-defn-vars  '(indent type ws-type-name classpec name ws-name-args args ws-args-post post))

;;; NOTE:
;;;    pre-command-hook  --> command executed before each command
;;;    last-input-char   --> last character input (useful to reimplement self-insert-command)


(defun c++-insert-char (&optional chr)
  (interactive)
  (or chr (setq chr last-input-char))

  "This function replaces self-insert-command to allow us to monitor editing of 
   C++ declarations and definitions.  If a character is being inserted into a
   C++ declaration (in a .h file), then the character is also inserted into
   the same relative portion of the C++ definition (in a .cc or .icc file). 
   Similarly, if currently editing a C++ definition (in a .cc or .icc file), then
   the character is also inserted into the same relative portion of the C++
   declaration (in the .h file).

   Note that in order to provide efficiency, there are some
   requirements on the format that declarations and definitions can
   take in order to benefit from this mirroring behavior.  In
   particular, declarations are assumed to occur only in header files
   (a file ending in c++-header-suffix), must be contained on a single
   line, and must match the regexp c++-decl-match (and note that this
   match ends in a ';', which means that specifications in a header
   file providing actual code (using a '{' instead of a ';') are
   ignored, as they should be.  Similarly, definitions occur only in
   source (or inline files) (i.e. files ending in c++-source-suffix or
   c++-inline-suffix) and must be defined on a single line and must
   match the regexp specified by c++-defn-match.  Note that this regexp
   OPTIONALLY ends with '{' because many people prefer to have their curly braces
   on the next line.  However, the regexp must match up to the end of the line in
   order to distinguish between methods that differ only in a 'const' method spec.
   "

  (cond
     ( (c++-in-header)
       (c++-parse-dd c++-decl-match c++-decl-vars 'decl chr c++-source-suffix c++-defn-vars)
     )
     ( (c++-in-source)
       (c++-parse-dd c++-defn-match c++-defn-vars 'defn chr c++-header-suffix c++-decl-vars)
     )
  )
)
(define-key my-map "cic" 'c++-insert-char)

(defun c++-in-header ()
   "Returns TRUE if current in a C++ header file (a file ending in c++-header-suffix)"
   (not (null (string-match (concat c++-header-suffix "$") (buffer-file-name))))
)

(defun c++-in-source ()
   "Returns TRUE if current in a C++ source file (a file ending in c++-source-suffix)"
   (not (null (string-match (concat c++-source-suffix "$") (buffer-file-name))))
)

(defun c++-in-inline ()
   "Returns TRUE if current in a C++ inline file (a file ending in c++-inline-suffix)"
   (not (null (string-match (concat c++-inline-suffix "$") (buffer-file-name))))
)

(defun c++-parse-dd (match vars type chr other-suffix other-vars)
   "Parse current line into the following components:
       indent        --> white-space at left margin
       type          --> a type specification
       ws-type-name  --> white-space between type and name
       name          --> name of method
       ws-name-args  --> white-space between method and args
       args          --> argument list (including paratheses)
       ws-args-post  --> white-space between args and post
       post          --> post text (either empty string or 'const')

    using the regexp specified by MATCH (which should have n subgroups) and
    forming an assoc-list mapping elements of VARS to subgroups (VARS should
    have n elements).  TYPE is one of 'decl or 'defn.

   If the current line does NOT match against MATCH, insert CHR at
   current point and do nothing more.  However, if the current line
   does match, we not only insert CHR at point, we also load up the
   file indicated by other-suffix and find the corresponding
   specification in that file, inserting CHR into the appropriate
   relative location.  This is not a completely trivial process
   because we want to allow for variations in whitespace usage between
   declarations and definitions, so one column in the declaration does
   not necessarily correspond to the same column in the definition
   (and in fact, the difference in column can vary depending on which
   portion of the specification is being modified).  
   "

   (let* ( (buf (current-buffer))
	   (pos (current-column))
           (line (current-line)) 
	   (other-file (concat (c++-current-file-base) other-suffix))
           (hash (c++-match-and-bind match line vars))
           (post (assoc-default 'post hash))
	   (other-match (format "^\\([ \t]*\\)\\(%s\\)\\([ \t]*\\)\\(%s[ \t]*::[ \t]*\\)\\(%s\\)\\([ \t]*\\)\\(%s\\)\\([ \t]*\\)\\(%s\\)"
                            (assoc-default 'type hash)
			    (c++-current-class)
                            (assoc-default 'name hash)
                            (assoc-default 'args hash)			 
                            (if (null post) "" post)
			 ))
         )

     (cond 
       ( (not (null hash))

	; We need to find corresponding specification in file matching
	; other-suffix and insert CHR at same relative location in that
        ; specification.
	(c++-load-file other-file)
	(goto-char (point-min))
        ; (message (format "%s" other-match))
        (message (format "%s" hash))
	(cond  
	  ( (re-search-forward other-match (point-max) t)

	    (let* ( (other-line (current-line))
		   (posloc (c++-analyze-hash hash vars))
		   (other-hash (c++-match-and-bind other-match other-line other-vars))
		   (other-posloc (c++-analyze-hash other-hash other-vars))
		   (other-pos (c++-find-matching-position hash posloc vars pos other-hash other-posloc other-vars type))
                 )
	       ; Using the current column position ('pos'), we look for the "section" of the specification
               ; that we are currently in (within posloc and hash).  We can find the column position of
	       ; the start of the proper section from posloc, and find the character offset within this
	       ; section be subtracting the start from the current position.  We then find the start of
	       ; the same section in other-posloc and add the same delta value to determine the location
	       ; in the other file to insert the character at.
	       (cond ( (not (null other-pos)) )
   	           (beginning-of-line)
	           (goto-char (+ (point) other-pos))
	           (insert chr)
               )
	       ; (end-of-line)
	       ;(insert (format "\n%s =\n %s\n\n%s =\n %s\n" hash posloc other-hash other-posloc))
             ) ; let
           ) ; found other-match
         ) ; inner cond  
       ) ; if not null hash
     ) ; outer cond 
     (set-buffer buf)
   )
   (insert chr)
  ; (other-window -1)
)

(defun c++-find-matching-position (hash posloc vars pos other-hash other-posloc other-vars type)
   "If POS is less than  posloc at (car vars), then (car vars) is the proper section.  We
    determine if the section has a corresponding other-section (we ignore white space sections,
    and some real sections don't have correspondences.  If it is not a real section, we return nil.
    If it is a real section, we find the corresponding section in other-posloc and from that, 
    the column position in other-file where we are to insert.
   "
   (cond 
      ( (null vars)
         nil
      )

      ( t
	(let* ( (v (car vars)) 
                (p (assoc-default v posloc))
             )
        )
      )
   )
)

(defun c++-analyze-hash (hash vars &optional length)
   "Returns an assoc of integers parallel to hash representing the cumulative lengths of
    the values (in HASH) associated with keys (in VARS).  For example, if VARS is '(a b c) and HASH is 
       '( (a . \"aaa\") (b . \"bbbb\") (c . \"cc\") )
    then the result is:
       ( (a . 3) (b . 7) (c . 9) )
   "
   (if (null length) (setq length 0))
   (if (null vars)
      nil
    ; else
      (let ( (len (+ length (length (assoc-default (car vars) hash)))) )
	 (cons (cons (car vars) len) (c++-analyze-hash hash (cdr vars) len))
      )
  )
)

(defun c++-load-file (file)
    "If FILE already exists in a currently displayed buffer, we simply set
     focus to that buffer.  If no window is currently displaying the buffer
     for file, we create such a window."
    (other-window 1)
    (find-file file)
)

(defun c++-match-and-bind (regexp str vars)
   "Apply REGEXP to STR and return an assoc list binding the first
    element of VARS to the first subgroup matched, the second element
    of VARS to the second subgroup matched, etc."
   (let ( (res (string-match regexp str)) )
      (cond ( (not (null res))
              (c++-match-and-bind-rec str vars 1)
            )
       )
   )
)

(defun c++-match-and-bind-rec (str vars index)
   "A recursive subfunction for 'c++-match-and-bind that forms the
    assoc mapping keys in VARS to match groups"
   (if (null vars)
      nil
    ; else
      (cons (cons (car vars) (match-string index str)) (c++-match-and-bind-rec str (cdr vars) (+ index 1)))
   )
)

(defun current-line ()
   (let* ( (p (point))
           (s (progn () (beginning-of-line) (point)))
           (e (progn () (end-of-line) (point)))
           (line (buffer-substring s e))
         )
       (goto-char p)
       line
   )
)

(defun my-capitalize (str) 
   ; (upcase-initials)  ; this probably won't exist in emacs (use next line if so)
   (concat (upcase (substring str 0 1)) (substring str 1))
)

(defun wmh-replace-in-string (str regexp newtext &optional literal) 
  "Replace all matches in STR for REGEXP with NEWTEXT string, 
 and returns the new string. 
Optional LITERAL non-nil means do a literal replacement. 
Otherwise treat `\\' in NEWTEXT as special: 
  `\\&' in NEWTEXT means substitute original matched text. 
  `\\N' means substitute what matched the Nth `\\(...\\)'. 
       If Nth parens didn't match, substitute nothing. 
  `\\\\' means insert one `\\'. 
  `\\u' means upcase the next character. 
  `\\l' means downcase the next character. 
  `\\U' means begin upcasing all following characters. 
  `\\L' means begin downcasing all following characters. 
  `\\E' means terminate the effect of any `\\U' or `\\L'.

  NOTE: Copied from the xemacs source code and modified to work
        for both xemacs and emacs so that c++.el works for both.
  " 
  ;;(check-argument-type 'stringp str) 
  ;;(check-argument-type 'stringp newtext) 
  (if (> (length str) 50) 
      (let ((cfs case-fold-search)) 
        (with-temp-buffer 
          (setq case-fold-search cfs) 
          (insert str) 
          (goto-char 1) 
          (while (re-search-forward regexp nil t) 
            (replace-match newtext t literal)) 
          (buffer-string))) 
  (let ((start 0) newstr) 
    (while (string-match regexp str start) 
      (setq newstr (replace-match newtext t literal str) 
            start (+ (match-end 0) (- (length newstr) (length str))) 
            str newstr)) 
    str))) 
