# ifndef OCRModule_h
# define OCRModule_h OCRModule_h

# include "ProcessComponent.h"
# include "SegmentComponent.h"
# include "IdentifyComponent.h"
# include "FormatComponent.h"

namespace Conjecture {
    // Forward Declarations
    class Env;

  /// \class OCRModule
  ///
  /// \brief AbstractFactory design pattern supporting the bundled
  /// creation of related Component algorithms.
  ///
  /// In order to make customization easy, Conjecture identifies the
  /// fundamental algorithms necessary to perform OCR. Each of these
  /// algorithms is formalized into a Component class hierarchy -
  /// different subclasses provide different implementations of the
  /// algorithm. A particular run of the OCR establishes which
  /// strategy to use for every one of its algorithms, and performs
  /// them in the appropriate sequence (the sequence in which
  /// algorithms are applied is also one of the algorithms).
  ///
  /// Since the specific strategy for solving one problem might depend
  /// on or require a specific strategy for solving some other
  /// pattern, we introduce an Abstract Factory design pattern. In
  /// general, an Abstract Factory is a collection of Factory Methods
  /// (which produce products).  An Abstract Factory allows one to
  /// enforce dependencies between individual products created.
  ///
  /// NOTE: The products being produced here are instances of a
  /// specific subclass of each Component hierarchy. Component is
  /// another design pattern, and normally the subclasses are
  /// Singletons (the classes represent algorithms, and are normally
  /// stateless, or at least requiring only a single state). However,
  /// I have not yet fully established whether using Singletons will
  /// pose problems for multi-threading or parallel/distributed
  /// algorithms, so for now I'm not making Component subclasses
  /// singletons. If we do decide to do so, it might be easier to just
  /// have the OCRModule intern objects than to put the
  /// Singleton design pattern into every single Component subclass.
  /// There will be many Component hierarchies, and potentially many
  /// subclasses in each hierarchy.

  class OCRModule {
  public:
    // Constructors/Destructors
    OCRModule();
    
    // Accessors
    //  - This class handles accessors for its algorithms differently than
    //    normal. The strategy is return non-const, and no 'xxxRef()'
    //    method is defined. This is because other classes always need
    //    non-const access to these algorithms in order to function.
    inline FormatComponent*          format()    const { return this->_format; }
    inline IdentifyComponent*        identify()  const { return this->_identify; }
    inline SegmentComponent*         segment()   const { return this->_segment; }
    inline ProcessComponent*         process()   const { return this->_process; }
    
    void init();
    
    // The Factory Methods that create every Component
    virtual ProcessComponent*        createProcessComponent()   const = 0;
    virtual SegmentComponent*        createSegmentComponent() const = 0;
    virtual IdentifyComponent*       createIdentifyComponent()  const = 0;
    virtual FormatComponent*         createFormatComponent()    const = 0;

    /**
     * Command-line processing
     *
     * This method provides module-specific command-line processing.
     *
     * The default implementation is a no-op.  Subclasses redefine as necessary.
     */
    virtual void processArgs(int argc, const char** argv, Env* env = NULL) { }

  protected:
    // Accessors
    inline void                     formatIs(FormatComponent* format) { this->_format = format; }
    inline void                     identifyIs(IdentifyComponent* identify) { this->_identify = identify; }
    inline void                     segmentIs(SegmentComponent* segment) { this->_segment = segment; }
    inline void                     processIs(ProcessComponent* process) { this->_process = process; }
    
    // Methods 
    
  private: 
    // Accessors

    // Methods 

    // State
    FormatComponent*                 _format;
    IdentifyComponent*               _identify;
    SegmentComponent*                _segment;
    ProcessComponent*                _process;
  };
};

# endif // OCRModule_h

