using namespace std;
# include <iostream>
# include "GlyphFeatures.h"
# include "Glyph.h"

// ####################################################################
// # Conjecture: An Extensible Optical Character Recognition Framework  #
// #                                                                  #
// # Copyright: Lesser GNU Public License (LGPL)                      #
// #                                                                  #
// # Overview:                                                        #
// #  - Conjecture is a collection of C++ classes, and an OCR engine    #
// #  - Conjecture is designed to allow customization at all levels     #
// #    and to encourage individuals to contribute incremental        #
// #    improvements in algorithms.                                   #
// #  - Overall design discussions can be found in                    #
// #      $SNROOT/docs/doxygen/html/index.html                        #
// #  - Conjecture is designed to interact with and build on other      #
// #    open-source OCR programs.                                     #
// ####################################################################
// Created 2006/06/10 by wade

// Include

namespace Conjecture {

    // CODE TANGLE #1: The indices here are tangled with the
    // values assigned to the 'Variant' enum in the header file!
    const char* GlyphFeatures::_LineTypes[] = {
        "LTUnknown", // 0
        "Base",      // 1
        "Ascender",  // 2
        "Descender", // 3
        "Total",     // 4
    };

    string GlyphFeatures::LineTypeStr(LineType lt) { return string(_LineTypes[lt]); }

    // CODE TANGLE #2: The indices here are tangled with the
    // values assigned to the 'Variant' enum in the header file!
    const char* GlyphFeatures::_Variants[] = {
        "",          // 0
        "Center",    // 1
        "Top",       // 2
        "Bottom",    // 3
        "Left",      // 4
        "Right",     // 5
        "",          // 6
        "",          // 7
        "Dominant",  // 8
        "Recessive", // 9
        "HalfLine",  // 10
        "Undef",     // 11
    };

    string GlyphFeatures::VariantStr(Variant v) { return string(_Variants[v]); }

    GlyphFeatures::GlyphFeatures(LineType lineType) {
        this->init(0, lineType);
    }

    GlyphFeatures::GlyphFeatures(wchar_t unicode, LineType lineType, const string& spec) {
        this->init(unicode, lineType);

        unsigned i = 0;
        //DEBUG: cerr << "FEATURES: " << spec << endl;
        while (i < spec.size()) {
            //DEBUG: cerr << "Procesing " << spec[i] << endl;
            switch (spec[i]) {
                case '':
                  // vertical Left/Dominant  line
                  this->addVertical(Left,Dominant);
                  break;
                  
              case 'L':
                // vertical Left/Recessive line
                this->addVertical(Left,Recessive);
                break;
                
              case 'l':
                // vertical Left/HalfLine  line
                this->addVertical(Left,HalfLine);
                break;
                
              case '':
                // vertical Right/Dominant  line
                this->addVertical(Right,Dominant);
                break;
                
              case 'R':
                // vertical Right/Recessive line
                this->addVertical(Right,Recessive);
                break;
                
              case 'r':
                // vertical Right/HalfLine  line
                this->addVertical(Right,HalfLine);
                break;
                
              case '':
                // vertical Center/Dominant  line
                this->addVertical(Center,Dominant);
                break;
                
              case 'C':
                // vertical Center/Recessive line
                this->addVertical(Center,Recessive);
                break;
                
              case 'c':
                // vertical Center/HalfLine  line
                this->addVertical(Center,HalfLine);
                break;

              case '':
                // horizontal Top/Dominant  line
                this->addHorizontal(Top,Dominant);
                break;

              case 'T':
                // horizontal Top/Recessive line
                this->addHorizontal(Top,Recessive);
                break;
                
              case 't':
                // horizontal Top/HalfLine  line
                this->addHorizontal(Top,HalfLine);
                break;

              case '':
                // horizontal Bottom/Dominant  line
                this->addHorizontal(Bottom,Dominant);
                break;
                
              case 'B':
                // horizontal Bottom/Recessive line
                this->addHorizontal(Bottom,Recessive);
                break;
                
              case 'b':
                // horizontal Bottom/HalfLine  line
                this->addHorizontal(Bottom,HalfLine);
                break;
                
              case '=':
                // horizontal Center/Dominant  line
                this->addHorizontal(Center,Dominant);
                break;
                
              case '-':
                // horizontal Center/Recessive line
                this->addHorizontal(Center,Recessive);
                break;
                
              case '~':
                // horizontal Center/HalfLine  line
                this->addHorizontal(Center,HalfLine);
                break;
                
              case '>':
                // Left cup
                this->addCup(Left);
                break;
                
              case '<':
                // Right cup
                this->addCup(Right);
                break;
                
              case '^':
                // Bottom cup
                this->addCup(Bottom);
                break;
                
              case 'v':
                // Top cup
                this->addCup(Top);
                break;
                
              case 'o':
                // Center hole
                this->addHole(Center);
                break;
                
              case 'e':
                // Top hole
                this->addHole(Top);
                break;
                
              case 'a':
                // Bottom hole
                this->addHole(Bottom);
                break;
                
              case '8':
                // Top hole and Bottom hole
                this->addHole(Top);
                this->addHole(Bottom);
                break;

              case 'y':
                this->ydisconnectIs(true);
                break;

              case 'x':
                this->xdisconnectIs(true);
                break;

              case 'Y':
                switch (spec[i+1]) {
                  case 'B': 
                    this->lineTypeIs(Base);
                    break;
                  case 'A': 
                    this->lineTypeIs(Ascender);
                    break;
                  case 'D': 
                    this->lineTypeIs(Descender);
                    break;
                  case 'T': 
                    this->lineTypeIs(Total);
                    break;
                  default:
                    cerr << "WARN: Illegal arg '" << spec[i+1] << "' to Y key (ignored)" << endl;
                }
                ++i;
                break;

              default:
                cerr << "WARN: Illegal key '" << spec[i] << "' ignored" << endl;
            }
            ++i;
        }
    }

    void GlyphFeatures::init(wchar_t unicode, LineType lineType) {
        this->unicodeIs(unicode);
        this->lineTypeIs(lineType);
        this->ydisconnectIs(false);
        this->xdisconnectIs(false);
    }
    
    // Interface
    void GlyphFeatures::addCup(Variant position) {
        this->cupsRef().push_back(position);
    }
    void GlyphFeatures::addHole(Variant position) {
        this->holesRef().push_back(position);
    }
    void GlyphFeatures::addVertical(Variant position, Variant length) {
        this->verticalRef().push_back(Variant2(position,length));
    }
    void GlyphFeatures::addHorizontal(Variant position, Variant length) {
        this->horizontalRef().push_back(Variant2(position,length));
    }

    GlyphFeatures::Variant GlyphFeatures::hasHorizontal(Variant position) const {
        // ridiculuous to have to search - much better implementation possible
        const Variant2Set& arr = this->horizontal();
        Variant result = Undef;
        for (Variant2Set::const_iterator it = arr.begin(); it != arr.end(); ++it) {
            if ( it->first == position ) {
                result = it->second;
                break;
            }
        }
        return result;
    }

    GlyphFeatures::Variant GlyphFeatures::hasVertical(Variant position) const {
        // ridiculuous to have to search - much better implementation possible
        const Variant2Set& arr = this->vertical();
        Variant result = Undef;
        for (Variant2Set::const_iterator it = arr.begin(); it != arr.end(); ++it) {
            if ( it->first == position ) {
                result = it->second;
                break;
            }
        }
        return result;
    }

    bool GlyphFeatures::hasHole(Variant position) const {
        // ridiculuous to have to search - much better implementation possible
        const VariantSet& arr = this->holes();
        bool result = false;
        for (VariantSet::const_iterator it = arr.begin(); it != arr.end(); ++it) {
            if ( *it == position ) {
                result = true;
                break;
            }
        }
        return result;
    }

    bool GlyphFeatures::hasCup(Variant position) const {
        // ridiculuous to have to search - much better implementation
        // possible
        const VariantSet& arr = this->cups();
        bool result = false;
        for (VariantSet::const_iterator it = arr.begin(); it != arr.end(); ++it) {
            if ( *it == position ) {
                result = true;
                break;
            }
        }
        return result;
    }

    GlyphFeatures::Confidence GlyphFeatures::matches(const GlyphFeatures& target) const {
        Confidence result = 0.0;

        int debug = 0;

        // ************************
        // Line Type Matching
        //  
        //  - A glyph can be categorized relative the Line it is contained
        //    within as being either 'Base', 'Ascender', 'Descender' or
        //    'Total'.  Each output character has an assigned line type.
        //
        //  - It is assumed unlikely that an Ascender will be
        //    misdiagnosied as a Descender or vice-versa, hence the
        //    stiff penalty.
        //
        //  - It is possible for a glyph to be categorized as Ascender
        //    when it is really Base if there is noise in the image.
        //    One pixel of unfiltered noise can cause this misdiagnosis,
        //    but lone pixels are filtered - how likely to misdiagnose?
        //
        //  - It is possible for a glyph to be categorized as Base
        //    when it is really Ascender if there is anti-noise in the
        //    image (pixels deleted). All pixels above the xheight
        //    would need to be anti-noised, which is usually unlikely.
        //    However, some glyphs have a minor presence above the
        //    baseline (i.e. t) and can easily be misdiagnosed.
	//    FUTURE FIX: May want to consider another lineType!!!
	//

	if ( this->lineType() == LTUnknown ) {
	    throw string("FatalException: receiver lineType has not been initialized");
	}
	if ( target.lineType() == LTUnknown ) {
	    throw string("FatalException: charset lineType has not been initialized");
	}


        Confidence typeconf = 0.0;
        switch( this->lineType() ) {
          case Base:
            switch (target.lineType()) {
              case Base      : typeconf =  1.00; break;
              case Ascender  : typeconf =  0.50; break;
              case Descender : typeconf =  0.30; break;
              case Total     : typeconf = -1.00; break;
              default: throw string("FATAL - matches fix!");
            }
            break;
          case Ascender:
            switch (target.lineType()) {
              case Base      : typeconf =  0.20; break;
              case Ascender  : typeconf =  1.00; break;
              case Descender : typeconf = -0.70; break;
              case Total     : typeconf =  0.20; break;
              default: throw string("FATAL - matches fix!");
            }
            break;
          case Descender:
            switch (target.lineType()) {
              case Base      : typeconf =  0.50; break;
              case Ascender  : typeconf = -0.70; break;
              case Descender : typeconf =  1.00; break;
              case Total     : typeconf =  0.20; break;
              default: throw string("FATAL - matches fix!");
            }
            break;
          case Total:
            switch (target.lineType()) {
              case Base      : typeconf = -1.00; break;
              case Ascender  : typeconf = -0.20; break;
              case Descender : typeconf = -0.50; break;
              case Total     : typeconf =  1.00; break;
              default: throw string("FATAL - matches fix!");
            }
            break;
          default:
            cerr << "FATAL in 'matches' - fix this!" << endl;
            exit(1);
        }
        result += typeconf * 0.30;

        // ************************
        // Line Stroke Matching.
        //  - there are 6 different kinds of lines.  We obtain
        //    a total line score by taking the average of the
        //    scores assigned to each of these 6 lines.
        Confidence lineconf = 0.0;

        if ( debug ) {  this->print(); target.print(); }

        lineconf += this->lineMatch(this->hasHorizontal(Top), target.hasHorizontal(Top));
        lineconf += this->lineMatch(this->hasHorizontal(Center), target.hasHorizontal(Center));
        lineconf += this->lineMatch(this->hasHorizontal(Bottom), target.hasHorizontal(Bottom));

        lineconf += this->lineMatch(this->hasVertical(Left), target.hasVertical(Left));
        lineconf += this->lineMatch(this->hasVertical(Center), target.hasVertical(Center));
        lineconf += this->lineMatch(this->hasVertical(Right), target.hasVertical(Right));

        lineconf /= 6;
        //DEBUG: cerr << "NOTE: lineconf is " << lineconf << endl;

        result += lineconf * 0.30;

        return result;
    }

    GlyphFeatures::Confidence GlyphFeatures::lineMatch(Variant mine, Variant target) const {
        Confidence result = 0.0;

        int debug = 0;

        // Use a matrix instead?? Or advancements that use conditional
        // logic justify this structure?
        switch (target) {
          case Dominant:
            switch(mine) {
              case Dominant  : result =  1.00; break;
              case Recessive : result =  0.90; break;
              case HalfLine  : result =  0.25; break;
              case Undef     : result = -1.00; break;
              default        : cerr << "ERROR: Invalid 'mine' variant type " << VariantStr(mine) << " passed to 'lineMatch'" << endl;
            }
            break;
          case Recessive:
            switch(mine) {
              case Dominant  : result =  1.00; break;
              case Recessive : result =  1.00; break;
              case HalfLine  : result =  0.80; break;
              case Undef     : result = -1.00; break;
              default        : cerr << "ERROR: Invalid 'mine' variant type " << VariantStr(mine) << " passed to 'lineMatch'" << endl;
            }
            break;
          case HalfLine:
            switch(mine) {
              case Dominant  : result = -0.10; break;
              case Recessive : result =  0.50; break;
              case HalfLine  : result =  1.00; break;
              case Undef     : result = -0.50; break;
              default        : cerr << "ERROR: Invalid 'mine' variant type " << VariantStr(mine) << " passed to 'lineMatch'" << endl;
            }
            break;
          case Undef:
            switch(mine) {
              case Dominant  : result = -1.00; break;
              case Recessive : result = -0.80; break;
              case HalfLine  : result = -0.50; break;
              case Undef     : result =  1.00; break;
              default        : cerr << "ERROR: Invalid 'mine' variant type " << VariantStr(mine) << " passed to 'lineMatch'" << endl;
            }
            break;
          default:
            cerr << "ERROR: Invalid target variant type " << VariantStr(target) << " passed to 'lineMatch'" << endl;
        }

        DBG(1, "NOTE:   " << VariantStr(mine) << " vs " << VariantStr(target) << " = " << result);
        
        return result;
    }

    ostream& GlyphFeatures::print(ostream& os, const string& indent) const {
        os << indent << "GlyphFeatures: ";
        if ( this->unicode() == 0 ) {
            os << "Unidentified glyph";
        } else {
            os << "Character " << (char)this->unicode();
        }
        os << endl;
        if ( this->lineType() != LTUnknown ) {
            os << indent << "  LineType   : " << LineTypeStr(this->lineType()) << endl;
        }

        // Holes
        const VariantSet& holes = this->holes();
        if ( holes.size() > 0 ) {
            os << indent << "  Holes      : (" << holes.size() << ")";
            for (VariantSet::const_iterator it = holes.begin(); it != holes.end(); ++it) {
                os << " " << VariantStr(*it);
            }
            os << endl;
        }

        // Cups
        const VariantSet& cups = this->cups();
        if ( cups.size() > 0 ) {
            os << indent << "  Cups       : (" << cups.size() << ")";
            for (VariantSet::const_iterator it = cups.begin(); it != cups.end(); ++it) {
                os << " " << VariantStr(*it);
            }
            os << endl;
        }

        // Vertical Lines
        const Variant2Set& vertical = this->vertical();
        if ( vertical.size() > 0 ) {
            os << indent << "  Vertical   : (" << vertical.size() << ")";
            for (Variant2Set::const_iterator it = vertical.begin(); it != vertical.end(); ++it) {
                os << " " << VariantStr(it->first) << "/" << VariantStr(it->second);
            }
            os << endl;
        }

        // Horizontal Lines
        const Variant2Set& horizontal = this->horizontal();
        if ( horizontal.size() > 0 ) {
            os << indent << "  Horizontal : (" << horizontal.size() << ")";
            for (Variant2Set::const_iterator it = horizontal.begin(); it != horizontal.end(); ++it) {
                os << " " << VariantStr(it->first) << "/" << VariantStr(it->second);
            }
            os << endl;
        }

        return os;
    }

    // The Unit Testing Method!!!
    void GlyphFeatures::Test(int argc, const char* argv[]) {
        GlyphFeatures feat(Ascender);
        feat.addHole(Top);
        feat.addCup(Bottom);
        feat.addVertical(Left,Recessive);
        feat.addVertical(Right,HalfLine);
        feat.addHorizontal(Top,Dominant);
        feat.print();

        GlyphFeatures feat2(0,Ascender, "Ae^Lr");
        feat2.print();
    }
};

