# ifndef OptIm_h
# define OptIm_h OptIm_h

// ####################################################################
// # Conjecture: An Extensible Optical Character Recognition Framework  #
// #                                                                  #
// # Copyright: Lesser GNU Public License (LGPL)                      #
// #                                                                  #
// # Overview:                                                        #
// #  - Conjecture is a collection of C++ classes, and an OCR engine    #
// #  - Conjecture is designed to allow customization at all levels     #
// #    and to encourage individuals to contribute incremental        #
// #    improvements in algorithms.                                   #
// #  - Overall design discussions can be found in                    #
// #      $SNROOT/docs/doxygen/html/index.html                        #
// #  - Conjecture is designed to interact with and build on other      #
// #    open-source OCR programs.                                     #
// ####################################################################

# include <vector>
# include "Root.h"  // parent
# include "PixelLine.h" // has-a

# include "GlyphFeatures.h"  // Variant


namespace Conjecture {
    class Image;
    class ImageArgs;
    
    // ##################################################################
    /// \class OptIm
    ///
    ///
    /// \brief An Optimized Image class (but only for small images)
    ///
    /// This class provides functionality and data for analyzing
    /// features of images of size 31x31 or less.
    ///
    /// IMPORTANT: There are two stages of initialization associated
    /// with this class (it partially violates the Conjecture
    /// programming mandate that all classes be fully initialized by
    /// their constructors).
    ///   - An OptIm contains both a collection of bits representing
    ///     pixels, and meta-data associated with these pixels.
    ///   - Furthermore, an OptIm maintains meta-data for both
    ///     rows of pixels and columns of pixels.
    ///   - The constructors only initialize the collection of
    ///     bits describing rows.  No metadata is computed, and
    ///     no column data is calculated.
    ///   - The 'updateMeta' method is used to have every PixelLine
    ///     compute its metadata, and to calculate (partially
    ///     redundanct) column data.
    ///   - This two-stage process exists for efficiency reasons.
    ///     There are various methods that access and modify only
    ///     the underlying raw pixel data, and it is better for
    ///     such modifications to be done before meta-data is
    ///     computed.  Incremental maintenance of the meta-data
    ///     during set/clear operations is expensive - better to
    ///     adjust the pixel data first, then compute metadata
    ///     once.
    ///
    /// One bit is used to represent each pixel, and is either "on"
    /// (black) or "off" (white). Additional "meta" information about
    /// each row and column is also maintained, including the number
    /// of holes and the number of black pixels. This information can
    /// be used to create very efficient feature-detection algorithms.
    ///
    /// The 'str' method is a useful debugging tool that shows the
    /// information maintained within an instance of this class in a
    /// very concise fashion.  An example of the output for the call:
    ///
    ///      optim->str('#','.',7)
    ///
    /// which generates:
    ///
    ///         0111110
    ///       0 #...... 1 
    ///       0 #...... 1 
    ///       0 #...... 1 
    ///       0 #...... 1 
    ///       1 #.###.. 4 
    ///       1 ##...#. 3 
    ///       1 #.....# 2 
    ///       1 #.....# 2 
    ///       0 .#####. 5 
    ///         8222222
    ///
    /// assuming that 'optim' is an Image* represents a 'b'. The
    /// numbers around the borders are interpreted as follows:
    ///    - left border   = number of holes in this row
    ///    - right border  = number of black pixels in this row
    ///    - top border    = number of holes in this column
    ///    - bottom border = number of black pixels in this column
    ///
    /// IMPORTANT: A "hole" in a line is a sequence of contiguous
    /// zero's (off or white bits) enclosed in 1's (on or black bits)
    /// on BOTH sides.
    ///
    /// FUTURE FIX: Optimization issue. This class currently stores
    /// partially redundant information because the PixelLines for
    /// columns are repeating pixel information already present in the
    /// row PixelLines. However, the column PixelLines are also
    /// providing non-redundant information (the metadata). In order
    /// to remove this redundancy, we will most likely need to throw
    /// away PixelLine and simply use this class as an atomic unit.
    /// Instead of a vector of PixelLines, we would maintain an array
    /// of pixel data, a parallel array of row hole counts, a parallel
    /// array of row black counts, a transposed array of column black
    /// counts, and a transposed array of column hole counts. This
    /// would improve both space utilization, and the efficiency of
    /// 'set' and 'clear' methods.
    // ##################################################################
    class OptIm : public Root {
        friend class PixFilt;

      public:
        // Types
        typedef std::vector<PixelLine> PixelMatrix;

        // Constructors/Destructors
        OptIm(const std::vector<unsigned int>& rows, u1 width = PixelLine::MaxPixels);
        OptIm(const Image* image, const ImageArgs& config);
        
        //static u1 BitImage(const Image* image, const ImageArgs& args, std::vector<u4>& rows);


        // **************
        // Accessors
        inline const u1 & width() const { return this->_width; }
        inline const PixelMatrix & cols() const { return this->_cols; }
        inline const PixelMatrix & rows() const { return this->_rows; }
        
        // **************
        // Input/Output
        
        // **************
        // Interface
        void updateMeta();

        // FUTURE FIX: These methods currently return 0 if the
        // instance isn't updated. Should we just be updateding the image
        // automatically? Or should we throw asserts that are removed
        // in production code?

        /// Number of black pixels in row 'i'
        int rblack(u1 i) const { return (this->updated() ? this->rows()[i].black() : 0); }
        /// Number of holes in row 'i'
        int rholes(u1 i) const { return (this->updated() ? this->rows()[i].holes() : 0); }
        /// Number of black pixels in column 'i'
        int cblack(u1 i) const { return (this->updated() ? this->cols()[i].black() : 0); }
        /// Number of holes in column 'i'
        int choles(u1 i) const { return (this->updated() ? this->cols()[i].holes() : 0); }

        u1 height() const { return this->rows().size(); } 

        /// Returns the pixel value at image coord <x,y> (origin at
        /// top-left corner of image).
        inline bool get(u2 x, u2 y) const { int w = this->width(); return this->rows()[y].get(w-1-x); }

        /// Create printable representation of this OptIm.
        ///
        /// See the class comments for more details.
        inline std::string str(char on = '#', char off = '.') const { return this->str("",0,on,off); }

        /// Create a string representation
        ///
        /// \param title  The title, placed above the column
        /// \param valign The number of rows to insert between title and start of data
        std::string str(const std::string& title, int valign = 0, char on = '#', char off = '.') const;

        /// Crop the image
        ///
        /// An image in which there are no black pixels along any
        /// border can be auto-cropped to a smaller image. More
        /// aggressive cropping can also be performed, in which
        /// continguous regions of "on" pixels that have a matching
        /// "on" pixel in the adjacent row/column are turned off (and
        /// the result potentially cropped).
        ///
        /// \param level If 0, basic cropping. If 1, perform
        /// contiguous region cropping in the one pixel thick
        /// boundary. If greater than 1, performs region cropping on
        /// the 'level'-pixel thick boundary.
        bool crop(int level = 0);
        bool crop(int topthick, int bottomthick, int leftthick, int rightthick);
        bool crop(int topthick, int bottomthick, int leftthick, int rightthick, int& topref, int& bottomref, int& leftref, int& rightref);

        /// Performs some serious work.  Feature detection.
        ///
        /// \param features out parameter - fields initialized by this method
        ///
        /// Note: Method is non-const because it auto-updates metadata
        /// if necessary. Should we throw exceptions in such cases
        /// instead?
        void identifyFeatures(GlyphFeatures& features);
        
        /// Obtain a specific const PixelLine by row. Negative indexes
        /// are relative to the end (-1 is last index, -2 is second
        /// last, etc.)
        inline const PixelLine& row(int i) const {
            if ( i < 0 ) { i += this->height(); }
            return this->rows()[i];
        }

        /// This method performs bottom serif removal on the image.
        ///
        /// NOTE: metadata is not needed for this method, and it is
        /// more efficient if this method is invoked before metadata
        /// is computed.
        bool removeBottomSerifs();

        /// This method performs top serif removal on the image.
        ///
        /// NOTE: metadata is not needed for this method, and it is
        /// more efficient if this method is invoked before metadata
        /// is computed.
        bool removeTopSerifs();

        static void test(int argc = 0, const char* argv[] = NULL);

        inline void print(ostream& os = cout) const { os << this->str() << endl; }
        
      protected:
        // **************
        // Accessors
        inline bool  updated() const                                 { return this->_updated; }
        inline void  updatedIs(const bool & updated)                   { this->_updated = updated; }
        inline void  widthIs(const u1 & width)                     { this->_width = width; }
        inline void  colsIs(const PixelMatrix & cols) { this->_cols  = cols;  }
        inline void  rowsIs(const PixelMatrix & rows) { this->_rows  = rows;  }
        
        // **************
        // Methods 
        std::string rawstr(const PixelMatrix& vec, char on = '1', char off = '0', unsigned char size = PixelLine::MaxPixels) const;

        /// Obtain a specific writeable PixelLine by row. Negative
        /// indexes are relative to the end (-1 is last index, -2 is
        /// second last, etc.)
        inline PixelLine& row(int i) {
            if ( i < 0 ) { i += this->height(); }
            return this->rowsRef()[i];
        }

        GlyphFeatures::Variant analyzeLines(u1 min, u2 max, const PixelMatrix& matrix, u1 dom, u1 rec, u1 halfn);
        
      private: 
        // **************
        // Accessors
        inline bool& updatedRef()                   { return this->_updated; }
        inline u1& widthRef()                       { return this->_width; }
        inline PixelMatrix& colsRef()  { return this->_cols;  }
        inline PixelMatrix& rowsRef()  { return this->_rows;  }
        
        // **************
        // Methods 
        
        // **************
        // State

        /// Internal boolean indication of whether the meta-information
        /// is up-to-date.
        bool                     _updated;

        /// The width in pixels
        ///
        /// Individual PixelLines can store up to 31 bits, but usually
        /// have fewer significant bits. They do not store their
        /// actual width to avoid redundancy - we can store the width
        /// once here.
        ///
        /// FUTURE FIX: is this necessary anymore? can't we just use
        /// cols.size()? Or do we need 'width' in order to compute
        /// 'cols'?
        u1                       _width;

        /// The row-based analysis
        PixelMatrix _rows;

        /// The column-based analysis
        PixelMatrix _cols;
    };
}

# endif // OptIm_h

