using namespace std;
# include <iostream>
# include <iomanip>
# include <sstream>
# include "PixFilt.h"
# include "OptIm.h"

// ####################################################################
// # Conjecture: An Extensible Optical Character Recognition Framework  #
// #                                                                  #
// # Copyright: Lesser GNU Public License (LGPL)                      #
// #                                                                  #
// # Overview:                                                        #
// #  - Conjecture is a collection of C++ classes, and an OCR engine    #
// #  - Conjecture is designed to allow customization at all levels     #
// #    and to encourage individuals to contribute incremental        #
// #    improvements in algorithms.                                   #
// #  - Overall design discussions can be found in                    #
// #      $SNROOT/docs/doxygen/html/index.html                        #
// #  - Conjecture is designed to interact with and build on other      #
// #    open-source OCR programs.                                     #
// ####################################################################

// Include

namespace Conjecture {

    PixFilt::PixFilt(u1 yoffset) {
        this->rowsIs(0);
        this->yoffsetIs(yoffset);
    }

    PixFilt::PixFilt(u1 yoffset, u1 mask1, int offset1) {
        this->rowsIs(0); // is adjusted by 'add'
        this->yoffsetIs(yoffset);
	this->add(mask1,offset1);
    }

    PixFilt::PixFilt(u1 yoffset, u1 mask1, int offset1, u1 mask2, int offset2) {
        this->rowsIs(0); // is adjusted by 'add'
        this->yoffsetIs(yoffset);
	this->add(mask1,offset1);
	this->add(mask2,offset2);
    }
    PixFilt::PixFilt(u1 yoffset, u1 mask1, int offset1, u1 mask2, int offset2, u1 mask3, int offset3) {
        this->rowsIs(0); // is adjusted by 'add'
        this->yoffsetIs(yoffset);
	this->add(mask1,offset1);
	this->add(mask2,offset2);
	this->add(mask3,offset3);
    }

    void PixFilt::add(u1 mask, int offset) {
        u1 r = this->rows();
        if ( r > 7 )                             { throw string("BoundsException: cannot handle more than 8 rows");             }
        if ( ( offset < -8 ) || ( offset > 7 ) ) { throw string("Bounds Exception: offset must be between -8 and 7 inclusive"); }
	
	u1 size = ffs(mask) + 1;

        int data = ( mask << 8 ) | ( ( offset + 8 ) << 4 ) | size;
        //DEBUG: cout << "mask=" << hex << (int)mask << " offset=" << dec << offset << " size=" << (int)size << endl;
        this->_data[r] = data;
        this->rowsIs(r+1);
        //DEBUG: cout << "mask=" << hex << (int)this->mask(r) << " offset=" << dec << this->offset(r) << " size=" << (int)this->size(r) << endl;
    }

    string PixFilt::str() const {
        stringstream s;
        for (int r = 0; r < this->rows(); ++r) {
            u1  mask   = this->mask(r);
            int offset = this->offset(r);
            u1  size   = this->size(r);
            s << string(8-offset, ' ');
            int i = size-1;
            int bit = (1<<i);
            while (i>=0) {
                s   << ((mask&bit) ? '1' : '0');
                bit >>= 1;
                --i;
            }
            s << endl;
        }
        return s.str();
    }

    bool PixFilt::match(const OptIm& im, u1 tx, u1 ty) {
        bool result = false;
        // Position <tx,ty> is the position within 'im', which has
        // <0,0> in the top left corner. However, this PixFilt has a
        // collection of row masks that have bit 0 on the right, not
        // the left. That is, <bx,by> = <imwidth-1-x,y> for any
        // image <x,y> coordinate.
        //
        // There are this->rows() rows in this mask, and row
        // this->yoffset() represents the "target" row, so we
        // start at row sy = ty - this->offset().
        int sy = ty - this->yoffset();

        // The end row: there are this->rows() rows.  start+size=end+1
        int ey = sy + this->rows() - 1;

	// The rows of data
	const vector<PixelLine>& rows = im.rows();

	u1 width = im.width();

	int height = rows.size();

        if ( (sy >= 0 ) && ( ey < height ) ) {
            // The current filter, when applied to row 'ty',
            // has enough space around it to apply.

            // We iterate over each row in this pixel filter
            for (int j = 0; j < this->rows(); ++j) {
                int y = sy + j;
                //DEBUG: cout << "ROW: " << y << endl;

                // We compute the start and end x positions (in image coords!)
                int sx = tx - this->offset(j);
                int ex = sx + this->size(j) - 1;
                //DEBUG: cout << "sx=" << sx << " ex=" << ex << endl;

                if ( ( sx < 0 ) || ( ex >= width ) ) {
                    // <tx,ty> is sufficient close to the left or
                    // right borders of the image that this filter
                    // cannot apply. NOTE: not *entirely* true (if ex
                    // > width but the mask has zeros in the overflow,
                    // it could still apply - fix this?)
                    result = false;
                    goto MEND;
                } else {
                    // We establish the mask to test against for this row.
                    u1 mask = this->mask(j);
                    // We obtain the image row
                    u4 row  = rows[y].pixels();

                    // We want to test 'mask' against pixel positions
                    // sx..ex of 'row' (keeping in mind that sx and ex
                    // are in image coords, not bit positions. We need
                    // to shift 'row' to the right a certain amount,
                    // and mask out high-order bits.
                    //   - if ex == imagewidth-1, we don't shift at
                    //     all.  If ex == imagewidth-2, we shift by
                    //     1, etc.  shift = (imagewidth-1)-ex.
                    //   - we know that the 'mask' will deal with the
                    //     bottom 8 bits, so after shifting row
                    //     we just look at its bottom 8 bits.
                    int shift = (width-1)-ex;
                    int chk   = (row >> shift) & 0xff;

                    //DEBUG: cout << "shift=" << shift << " mask=" << hex << (int)mask << " row=" << (int)row << dec << endl;

                    if ( (chk&mask) != mask ) {
                        // This row does NOT match the bit pattern, so
                        // we terminate.
                        result = false;
                        goto MEND;
                    }
                }
            }
            // If we've reached this point, this filter matches!
            result = true;

        } else {
            // <tx,ty> is sufficiently close to the top or bottom
            // border that this filter cannot apply (assuming that
            // no mask in _data is 0x0, of course).
            result = false;
        }

      MEND:
        return result;
    }

    bool PixFilt::match(const vector<u4>& rows, u1 width, u1 tx, u1 ty) {
	// FUTURE FIX: SHOULD THIS CODE BE REMOVED??? It is highly
	// redundant wrt match(OptIm), but is useful for other algorithms
	// that do not want to rely on OptIm!!

        bool result = false;
        // Position <tx,ty> is the position within 'im', which has
        // <0,0> in the top left corner. However, this PixFilt has a
        // collection of row masks that have bit 0 on the right, not
        // the left. That is, <bx,by> = <imwidth-1-x,y> for any
        // image <x,y> coordinate.
        //
        // There are this->rows() rows in this mask, and row
        // this->yoffset() represents the "target" row, so we
        // start at row sy = ty - this->offset().
        int sy = ty - this->yoffset();

        // The end row: there are this->rows() rows.  start+size=end+1
        int ey = sy + this->rows() - 1;

	int height = rows.size();

        if ( (sy >= 0 ) && ( ey < height ) ) {
            // The current filter, when applied to row 'ty',
            // has enough space around it to apply.

            // We iterate over each row in this pixel filter
            for (int j = 0; j < this->rows(); ++j) {
                int y = sy + j;
                //DEBUG: cout << "ROW: " << y << endl;

                // We compute the start and end x positions (in image coords!)
                int sx = tx - this->offset(j);
                int ex = sx + this->size(j) - 1;
                //DEBUG: cout << "sx=" << sx << " ex=" << ex << endl;

                if ( ( sx < 0 ) || ( ex >= width ) ) {
                    // <tx,ty> is sufficient close to the left or
                    // right borders of the image that this filter
                    // cannot apply. NOTE: not *entirely* true (if ex
                    // > width but the mask has zeros in the overflow,
                    // it could still apply - fix this?)
                    result = false;
                    goto MEND;
                } else {
                    // We establish the mask to test against for this row.
                    u1 mask = this->mask(j);
                    // We obtain the image row
                    u4 row  = rows[y];

                    // We want to test 'mask' against pixel positions
                    // sx..ex of 'row' (keeping in mind that sx and ex
                    // are in image coords, not bit positions. We need
                    // to shift 'row' to the right a certain amount,
                    // and mask out high-order bits.
                    //   - if ex == imagewidth-1, we don't shift at
                    //     all.  If ex == imagewidth-2, we shift by
                    //     1, etc.  shift = (imagewidth-1)-ex.
                    //   - we know that the 'mask' will deal with the
                    //     bottom 8 bits, so after shifting row
                    //     we just look at its bottom 8 bits.
                    int shift = (width-1)-ex;
                    int chk   = (row >> shift) & 0xff;

                    //DEBUG: cout << "shift=" << shift << " mask=" << hex << (int)mask << " row=" << (int)row << dec << endl;

                    if ( (chk&mask) != mask ) {
                        // This row does NOT match the bit pattern, so
                        // we terminate.
                        result = false;
                        goto MEND;
                    }
                }
            }
            // If we've reached this point, this filter matches!
            result = true;

        } else {
            // <tx,ty> is sufficiently close to the top or bottom
            // border that this filter cannot apply (assuming that
            // no mask in _data is 0x0, of course).
            result = false;
        }

      MEND:
        return result;
    }

    int PixFilt::filter(OptIm& im, bool set) {
	// CODE TANGLE: This code is very similar to the
	// filter(vector) method above!  Changes to either
	// should occur in both.

        // We have an image with given width and height.
        int w = im.width();
        int h = im.height();

        // We have an pixel filter with given yoffset and
        // vertical span (ysize)
        int yoffset = this->yoffset();
        int ysize   = this->rows();
        
        // When considering possible target positions:
        //  - can ignore the top 'yoffset' rows (cannot apply)
        //  - can ignore the bottom 'ysize-yoffset-1' rows (cannot apply)
        //  - since the x positions vary from row to row, we cannot
        //    necessarily ignore them here (the 'match' method deals
        //    with horizontal overflow properly).
	//  - only need to consider pixels that are currently cleared,
	//    (for now, PixFilt is always about setting bits).

	vector<PixelLine>& rows = im.rowsRef();

	int count = 0;
	//DEBUG: cout << "w=" << w << " h=" << h << endl;
	for (int y = yoffset; y < h-((ysize-yoffset-1)); ++y) {
	    for (int x = 0; x < w; ++x) {
		int bit = w-1-x;
		if ( rows[y].get(bit) != set ) {
		    // We have a location that is off if we are turning
		    // on, or that is on if we are turning off.
		    if ( this->match(im,x,y) ) {
			//DEBUG: cout << "<x,y> = <" << x << "," << y << ">" << endl;
			if (set) {
			    // Turn bit 'bit' on.
			    rows[y].rawset(bit);
			} else {
			    // Turn bit 'bit' off.
			    rows[y].rawclear(bit);
			}
			++count;
		    }
		}
	    }
	}
	return count;
    }

    int PixFilt::filter(vector<u4>& rows, u1 width, bool set) {
	// CODE TANGLE: This code is very similar to the
	// filter(OptIm) method above!  Changes to either
	// should occur in both.
	//
	// This method is not used in WmhAlgorithms code, but
	// is provided in case it is useful to other algorithms
	// that do not use OptIm instances.

        // We have an image with given width and height.
        int w = width;
        int h = rows.size();

        // We have an pixel filter with given yoffset and
        // vertical span (ysize)
        int yoffset = this->yoffset();
        int ysize   = this->rows();
        
        // When considering possible target positions:
        //  - can ignore the top 'yoffset' rows (cannot apply)
        //  - can ignore the bottom 'ysize-yoffset-1' rows (cannot apply)
        //  - since the x positions vary from row to row, we cannot
        //    necessarily ignore them here (the 'match' method deals
        //    with horizontal overflow properly).
	//  - only need to consider pixels that are currently cleared,
	//    (for now, PixFilt is always about setting bits).

	int count = 0;
	//DEBUG: cout << "w=" << w << " h=" << h << endl;
	for (int y = yoffset; y < h-((ysize-yoffset-1)); ++y) {
	    for (int x = 0; x < w; ++x) {
		int bit = w-1-x;
		int mask = 1<<bit;
		if ( (rows[y] & mask) == (!set)) {
		    // We have a location that is off if we are turning
		    // on, or that is on if we are turning off.
		    if ( this->match(rows,width,x,y) ) {
			//DEBUG: cout << "<x,y> = <" << x << "," << y << ">" << endl;
			if (set) {
			    // Turn bit 'bit' on.
			    rows[y] |= mask;
			} else {
			    // Turn bit 'bit' off.
			    rows[y] &= ~mask;
			}
			++count;
		    }
		}
	    }
	}
	return count;
    }

    int PixFilt::ffs(u1 val) {
	int result = -1;
	if ( val ) {
	    if ( val & 0xf0 ) {
		if ( val & 0xc0 ) {
		    result = (val&0x80) ? 7 : 6;
		} else {
		    result = (val&0x20) ? 5 : 4;
		}
	    } else {
		if ( val & 0xc ) {
		    result = (val&0x8) ? 3 : 2;
		} else {
		    result = (val&0x2) ? 1 : 0;
		}
	    }
	}
	return result;
    }
    
    // Interface
    void PixFilt::test(int argc, const char* argv[]) {

	//for (int k = 0; k < 256; ++k) { cout << setw(3) << k << " = " << ffs(k) << endl; }

	// Create a vector describing the following image
        //    #......
        //    #......
        //    #......
        //    #......
        //    #.###..
        //    ##...#.
        //    #...#.#
        //    #.....#
        //    .#####.
        vector<unsigned int> rows;
        rows.push_back(0x40); // 0
        rows.push_back(0x40); // 1
        rows.push_back(0x40); // 2
        rows.push_back(0x40); // 3
        rows.push_back(0x5d); // 4
        rows.push_back(0x62); // 5
        rows.push_back(0x45); // 6
        rows.push_back(0x41); // 7
        rows.push_back(0x3e); // 8
	int width = 7;

	// Now create an OptIm and print it
        OptIm im(rows, width);
        cout << im.str('#','.');

	// Now create a filter:
	//    ?1       ?1
	//    10  -->  11
	//    ?1       ?1
	cout << "Right Pixel" << endl;
        // PixFilt f(1);
	//  f.add(0x1, 0);
	//  f.add(0x2, 1);
	//  f.add(0x1, 0);
	PixFilt f(1,0x1,0,0x2,1,0x1,0);
        cout << f.str();

	// And a second filter:
	//    1?       1?
	//    01  -->  11
	//    1?       1?
	cout << "Left Pixel" << endl;
        //PixFilt f2(1);
        //f2.add(0x1, 0);
        //f2.add(0x1,-1);
        //f2.add(0x1, 0);
        PixFilt f2(1,0x1,0,0x1,-1,0x1,0);
        cout << f2.str();

	// And a third filter:
	//   101  --> 111
	cout << "Horizontal Pixel" << endl;
	//PixFilt g(0);
	//g.add(0x5,1);
	PixFilt g(0,0x5,1);
	cout << g.str();

	// Now apply the filters to the pixel rows
	int c = f.filter(rows,width);
	cout << "replaced " << c << " pixels" << endl;
	int c2 = f2.filter(rows,width);
	cout << "replaced " << c2 << " pixels" << endl;
	int c3 = g.filter(rows,width);
	cout << "replaced " << c3 << " pixels" << endl;

	// Create a new OptIm for the modified data.
        OptIm im2(rows, width);
        cout << im2.str('#','.');

	// u1 width = im.width();
	
	// cerr << ( ( f.match(rows, width, 6, 5) ) ? "MATCHES" : "FAILS" ) << endl;
	// cerr << ( ( f.match(rows, width, 2, 5) ) ? "MATCHES" : "FAILS" ) << endl;
	
	// PixFilt g(0);
	// g.add(0x5,1,3);
	// cout << g.str();
	// cerr << (( g.match(rows,width,1,4) ) ? "MATCHES" : "FAILS") << endl;
	// cerr << (( g.match(rows,width,5,4) ) ? "MATCHES" : "FAILS") << endl;
	// cerr << (( g.match(rows,width,4,4) ) ? "MATCHES" : "FAILS") << endl;

	
    }

};

