# ifndef PixFilt_h
# define PixFilt_h PixFilt_h

// ####################################################################
// # Conjecture: An Extensible Optical Character Recognition Framework  #
// #                                                                  #
// # Copyright: Lesser GNU Public License (LGPL)                      #
// #                                                                  #
// # Overview:                                                        #
// #  - Conjecture is a collection of C++ classes, and an OCR engine    #
// #  - Conjecture is designed to allow customization at all levels     #
// #    and to encourage individuals to contribute incremental        #
// #    improvements in algorithms.                                   #
// #  - Overall design discussions can be found in                    #
// #      $SNROOT/docs/doxygen/html/index.html                        #
// #  - Conjecture is designed to interact with and build on other      #
// #    open-source OCR programs.                                     #
// ####################################################################

# include "Root.h"  // parent
# include <vector>

namespace Conjecture {
    class OptIm;
    
    // ##################################################################
    /// \class PixFilt
    ///
    /// \brief Provides a generalized pixel filtering mechanism.
    ///
    /// This class acts as a 2d "bit-based regular expression",
    /// bitexp. For example, suppose we want to determine whether a
    /// particular pixel <tx,ty> within an image matches the following
    /// bitexp:
    ///
    ///        ?1?
    ///        10?
    ///        ?1?
    ///
    /// In such a situation, we may want to switch the 0 to a 1.
    /// Instances of this class can describe this kind of matching and
    /// can establish whether it is true for a particular pixel
    /// position within an image.
    ///
    /// To explain the rather esoteric data-structures, we'll
    /// describe the internal state needed to describe the above
    /// mask.
    ///   - it consists of 3 consecutive row-specific tests
    ///   - the first row test involves verifying that a one-pixel
    ///     sequence matches 1.
    ///   - the second row test involves verifying that a two-pixel
    ///     sequence matches 10.  This match is offset by one
    ///     relative to the first one.
    ///   - the third row test involves verifying that a one-pixel
    ///     sequence matches 1.  Same offset as first row.
    ///
    /// The state:
    ///    _rows     = 3
    ///    _yoffset  = 1
    ///    _data[0]  = (0x1  << 8) | ( ( 0+8)<<4 ) | 1
    ///    _data[1]  = (0x10 << 8) | ( (-1+8)<<4 ) | 1
    ///    _data[2]  = (0x1  << 8) | ( ( 0+8)<<4 ) | 1
    ///
    /// SUBTLETIES
    /// ----------
    ///  - There is an unfortute difference between the <x,y>
    ///    coordinate system defined on images, which places <0,0> in
    ///    the top-left corner, and the <bx,by> bit position define on
    ///    bit vectors, which places bit 0 on the right, *not* the
    ///    left.
    ///  - If 'tx' is the target x position, it will be very rare
    ///    for any of the row masks to have a left position that
    ///    is to the right of 'tx' (masks usually center 'tx' within
    ///    them, so the left most part of the mask is usually to
    ///    the left of 'tx'.  We make the common case positive, so
    ///    the 'offset' value indicates the number of pixels to shift
    ///    left (positive going left, negative going right). 

    // ##################################################################
    class PixFilt : public Root {
      public:
        // Constructors/Destructors
        PixFilt(u1 yoffset);
	PixFilt::PixFilt(u1 yoffset, u1 mask1, int offset1);
	PixFilt::PixFilt(u1 yoffset, u1 mask1, int offset1, u1 mask2, int offset2);
	PixFilt::PixFilt(u1 yoffset, u1 mask1, int offset1, u1 mask2, int offset2, u1 mask3, int offset3);
        
        // **************
        // Accessors
        inline const u2*       data()    const { return this->_data;    }
        inline const u1 &      rows()    const { return this->_rows;    }
        inline const u1 &      yoffset() const { return this->_yoffset; }
        
        // **************
        // Input/Output
        
        // **************
        // Interface
        void add(u1 mask, int offset);

        std::string str() const;

        u1 mask(u1 index)    const { return ((this->_data[index]) >> 8);       }
        int offset(u1 index) const { return ((this->_data[index] >> 4)&0xf)-8; }
        u1 size(u1 index)    const { return ((this->_data[index]) & 0xf);      }

        /// Returns true if this PixFilt matches against <x,y> of 'im'
        bool match(const std::vector<u4>& rows, u1 width, u1 x, u1 y);
        bool match(const OptIm& im, u1 x, u1 y);

        /// Apply this filter to the entire image
	///
	/// \param set True means turning on pixels that are off.
	///            False means turning off pixels that are on.
	///
	/// Returns the number of pixels changed.
        int filter(std::vector<u4>& im, u1 width, bool set = true);
        int filter(OptIm& im, bool set = true);

        static void test(int argc = 0, const char* argv[] = NULL);
        
      protected:
        // **************
        // Accessors
        inline void rowsIs(const u1 & rows)       { this->_rows = rows; }
        inline void yoffsetIs(const u1 & yoffset) { this->_yoffset = yoffset; }
        
        // **************
        // Methods 

	/// find-first-set: returns the bit position of the left-most on bit within 'val.
	/// 
	/// Bit 0 is on far right.  The value 0x5 would return 2 because
	/// 0x5 in binary is 101 (left most 1 is bit 2).
	/// The value '0x21' would return 5 because 0x21 is binary
	/// 100001 (left most 1 is bit 5).
	static int ffs(u1 val);
        
      private: 
        // **************
        // Accessors
        inline u1  &  rowsRef()    { return this->_rows;    }
        inline u1  &  yoffsetRef() { return this->_yoffset; }
        
        // **************
        // Methods 
        
        // **************
        // State

        /// Number of rows in the filter
        u1              _rows;

        /// The y position of the target pixel, relative to
        /// the matrix defined by this filter.
        u1              _yoffset;

        /// The raw data
        ///
        /// Each u2 entry encodes three pieces of information
        ///   4 bits  = size   (number of meaningful bits in mask)
        ///   4 bits  = offset (signed offset (-8 to 7))
        ///   8 bits  = mask   (1 to 8 bit mask)
        u2              _data[8];
    };
}

# endif // PixFilt_h

