using namespace std;
# include <iostream>
# include <sstream>
# include <iomanip>
# include "PixelLine.h"

// ####################################################################
// # Conjecture: An Extensible Optical Character Recognition Framework  #
// #                                                                  #
// # Copyright: Lesser GNU Public License (LGPL)                      #
// #                                                                  #
// # Overview:                                                        #
// #  - Conjecture is a collection of C++ classes, and an OCR engine    #
// #  - Conjecture is designed to allow customization at all levels     #
// #    and to encourage individuals to contribute incremental        #
// #    improvements in algorithms.                                   #
// #  - Overall design discussions can be found in                    #
// #      $SNROOT/docs/doxygen/html/index.html                        #
// #  - Conjecture is designed to interact with and build on other      #
// #    open-source OCR programs.                                     #
// ####################################################################

// Include

namespace Conjecture {

    PixelLine::PixelLine(u4 p) {
	// Only allow 31 pixels
	p &= 0x7fffffff;
	
	// Store the pixels.  Pixel 0 is right-most pixel.
	this->pixelsIs(p);

	// Meta data not yet computed
	this->updatedIs(false);
	this->blackIs(0);
	this->holesIs(0);
    }
    
    void PixelLine::updateMeta() {
	u4 p = this->pixels();
	
	// Temp vars
	u1 black    = 0;
	u1 flips    = 0;
	bool laston = false;
	
	// Iterate: check low-order bit, shift right one.
	while (p) {
	    if ( p&0x1 ) {
		// Pixel is black
		++black;
		if (!laston) { ++flips; }
		laston = true;
	    } else {
		// Pixel is white
		if (laston) { ++flips; }
		laston = false;
	    }
	    // shift right one
	    p >>= 1;
	}
	
	// The number of holes is half the number of flips:
	u1 holes = int(flips/2);
	this->blackIs(black);
	this->holesIs(holes);

	this->updatedIs(true);
    }

    // Interface
    string PixelLine::str(char on, char off, u1 size) const {
	stringstream s;
	if ( this->updated() ) {
	    s << setw(2) << (int)this->holes() << " ";
	} else {
	    s << setw(3) << " ";
	}

	for (int i = size-1; i >= 0; --i) {
	    if ( this->get(i) ) {
		s << on;
	    } else {
		s << off;
	    }
	}
	if ( this->updated() ) {
	    s << " " << left << setw(2) << (int)this->black();
	}

	return s.str();
    }

    void PixelLine::set(unsigned i)  {
	// establish current value.
	bool on = this->get(i);

	if (!this->updated()) { this->updateMeta(); }
	
	if (!on) {
	    // We are switching from white to black
	    //  - increment black count
	    this->blackIs(this->black()+1);
	    //  - if we have '1X1' (X=0), reduce hole count by 1
	    //  - if we have '1X01' or '10X1', no affect on hole count
	    //  - if we have '10X01' or '00X01' or '10X00', increase hole count by 1
	    bool left  = (i<MaxPixels) ? this->get(i+1) : false;
	    bool right = (i>0)  ? this->get(i-1) : false;
	    int adj = 0;
	    if ( left && right ) { adj = -1; }
	    else if ( (!left)&&(!right) ) { adj = 1; }
	    if ( adj ) { this->holesIs(this->holes()+adj); }
	    
	    // Make this pixel black
	    this->rawset(i);
	}
    }
    
    void PixelLine::clear(unsigned i)  {
	// establish current value.
	bool on = this->get(i);
	
	if (!this->updated()) { this->updateMeta(); }

	if (on) {
	    // We are switching from black to white
	    //  - increment black count
	    this->blackIs(this->black()-1);
	    //  - if we have '1X1', increase hole count by 1
	    //  - if we have '0X1' or '1X0', no hole affect
	    //  - if we have '10X01 or '00X01' or '10X00', decrease hole count by 1
	    bool left  = (i<MaxPixels) ? this->get(i+1) : false;
	    bool right = (i>0)  ? this->get(i-1) : false;
	    int adj = 0;
	    if ( left && right ) { adj = 1; }
	    else if ( (!left)&&(!right) ) { adj = -1; }
	    if ( adj ) { this->holesIs(this->holes()+adj); }
	    
	    // Make this pixel white
	    this->rawclear(i);
	}
    }

    void PixelLine::assign(const PixelLine& other, u1 rightdel) {
	this->pixelsIs(other.pixels()>>rightdel);
	this->holesIs(other.holes());
	this->blackIs(other.black());
	this->updatedIs(other.updated());
    }
    
    void PixelLine::test(int argc, const char* argv[]) {
	// 0100010010110011
	PixelLine p1(0x44b3);
	cout << p1.str() << endl;
	p1.updateMeta();
	cout << p1.str() << endl;

	p1.set(2);
	cout << p1.str() << endl;
	p1.set(3);
	cout << p1.str() << endl;
	p1.set(3);
	cout << p1.str() << endl;
	p1.clear(3);
	cout << p1.str() << endl;
	p1.clear(3);
	cout << p1.str() << endl;
	p1.clear(2);
	cout << p1.str() << endl;
	
	PixelLine p2(0x15);
	cout << p2.str('*','.') << endl;
	p2.set(1);
	cout << p2.str('*','.') << endl;
	p2.set(3);
	cout << p2.str('*','.') << endl;
	p2.set(3);
	cout << p2.str('*','.') << endl;
	p2.clear(3);
	cout << p2.str('*','.') << endl;
	p2.clear(3);
	cout << p2.str('*','.') << endl;
	p2.clear(1);
	cout << p2.str('*','.') << endl;
	
	PixelLine p3(0x5);
	cout << p3.str('#',' ') << endl;
	p3.set(2);
	cout << p3.str('#',' ') << endl;
	p3.set(1);
	cout << p3.str('#',' ') << endl;
	p3.clear(1);
	cout << p3.str('#',' ') << endl;
	p3.clear(2);
	cout << p3.str('#',' ') << endl;
	
	PixelLine p4(0xa);
	cout << p4.str() << endl;
	
	PixelLine p5(0x80);
	cout << p5.str() << endl;
    }
};

