# ifndef OptIm15_h
# define OptIm15_h OptIm15_h

// ####################################################################
// # Conjecture: An Extensible Optical Character Recognition Framework  #
// #                                                                  #
// # Copyright: GNU Public License                                    #
// #                                                                  #
// # Overview:                                                        #
// #  - Conjecture is a collection of C++ classes, and an OCR engine    #
// #  - Conjecture is designed to allow customization at all levels     #
// #    and to encourage individuals to contribute incremental        #
// #    improvements in algorithms.                                   #
// #  - Overall design discussions can be found in                    #
// #      $SNROOT/docs/doxygen/html/index.html                        #
// #  - Conjecture is designed to interact with and build on other      #
// #    open-source OCR programs.                                     #
// ####################################################################

# include <vector>
# include "Root.h"  // parent
# include "PixelLine15.h" // has-a

namespace Conjecture {
  class Image;
  class ImageArgs;

  // ##################################################################
  /// \class OptIm15
  ///
  /// \brief An Optimized Image class (but only for very small images)
  ///
  /// This class provides functionality and data for analyzing
  /// features of tiny images (size 16x16 or less).
  ///
  /// One bit is used to represent each pixel, and is either on
  /// (black) or off (white). Information about each row and column is
  /// maintained, including the number of holes and the number of
  /// black pixels. This information will hopefully allow for the
  /// implementation of efficient methods related to feature
  /// extraction.
  ///
  /// The 'str' method is a useful debugging tool that shows the
  /// information maintained within an instance of this class in a
  /// very concise fashion.  An example of the output for the call:
  ///     optim->str('#','.',7)
  /// will be
  ///         0111110
  ///       0 #...... 1 
  ///       0 #...... 1 
  ///       0 #...... 1 
  ///       0 #...... 1 
  ///       1 #.###.. 4 
  ///       1 ##...#. 3 
  ///       1 #.....# 2 
  ///       1 #.....# 2 
  ///       0 .#####. 5 
  ///         8222222
  /// assuming that 'optim' represents the 'b' shown.  The numbers
  /// around the borders are interpreted as follows:
  ///    - left border   = number of holes in this row
  ///    - right border  = number of black pixels in this row
  ///    - top border    = number of holes in this column
  ///    - bottom border = number of black pixels in this column
  ///
  /// IMPORTANT: A "hole" in a line is a sequence of contiguous zero's
  /// (off or white bits) enclosed in 1's (on or black bits) on both
  /// ends.

  // ##################################################################
  class OptIm15 : public Root {
  public:
    // Constructors/Destructors
    OptIm15(const Image* image, const ImageArgs& config);
    OptIm15(const std::vector<unsigned int>& rows, unsigned char width = 15);

    // **************
    // Accessors
    inline const uint8_t & width() const { return this->_width; }
    inline const std::vector<PixelLine15> & cols() const { return this->_cols; }
    inline const std::vector<PixelLine15> & rows() const { return this->_rows; }
    
    // **************
    // Input/Output
    
    // **************
    // Interface

    void init(const std::vector<unsigned int>& rows, unsigned char width);

    /// Number of black pixels in row 'i'
    int rblack(unsigned char i) const { return this->rows()[i].black(); }
    /// Number of holes in row 'i'
    int rholes(unsigned char i) const { return this->rows()[i].holes(); }
    /// Number of black pixels in column 'i'
    int cblack(unsigned char i) const { return this->cols()[i].black(); }
    /// Number of holes in column 'i'
    int choles(unsigned char i) const { return this->cols()[i].holes(); }

    /// Create printable representation of this OptIm15.
    ///
    /// See the class comments for more details.
    std::string str(char on = '#', char off = '.') const; // , unsigned char size = 15) const;

    static void test(int argc = 0, const char* argv[] = NULL);

  protected:
    // **************
    // Accessors
    inline void            widthIs(const uint8_t & width) { this->_width = width; }
    inline void            colsIs(const std::vector<PixelLine15> & cols) { this->_cols = cols; }
    inline void            rowsIs(const std::vector<PixelLine15> & rows) { this->_rows = rows; }
    
    // **************
    // Methods 
    std::string rawstr(const std::vector<PixelLine15>& vec, char on = '1', char off = '0', unsigned char size = 15) const;    
    
  private: 
    // **************
    // Accessors
    inline uint8_t&                widthRef() { return this->_width; }
    inline std::vector<PixelLine15>& colsRef()  { return this->_cols; }
    inline std::vector<PixelLine15>& rowsRef()  { return this->_rows; }
    
    // **************
    // Methods 

    // **************
    // State
    uint8_t                _width;
    std::vector<PixelLine15> _cols;
    std::vector<PixelLine15> _rows;
  };
}

# endif // OptIm15_h

