# ifndef Conjecture_Element_h
# define Conjecture_Element_h Conjecture_Element_h

// C++ Dependencies
# include "Root.h"        // parent
# include <list>          // field 'parts'
# include <vector>        // method 'allGlyphs'
# include <iostream>      // cerr
# include "Coord.h"       // Coord
# include "ImageArgs.h"

// Declaration
namespace Conjecture {
    // Forward declarations
    class Page;
    class Region;
    class Line;
    class Word;
    class Glyph;
    class Image;
    
    // Types of Elements
    typedef enum ElementType {
        ELEMENT_PAGE,
        ELEMENT_REGION,
        ELEMENT_LINE,
        ELEMENT_WORD,
        ELEMENT_GLYPH
    };
    
    /********************************************************************
     * \class Element
     *
     * \brief Abstract superclass of a collection of classes
     * representing a part-whole decomposition of a graphic image into
     * smaller and smaller semantic units.
     *
     * The Element class is at the root of a critical hierarchy of
     * subclasses. Certain subclasses (Page and Glyph) reprsents the
     * most important classes in the implementation, and much of their
     * functionality is inherited from Element. Other subclasses
     * (Region, Line and Word) provide "refinements" but aren't
     * strictly necessary to the proper execution of the code.
     *
     * An Element consists of state and functionality supporting a
     * hierarchial decomposition of larger parts into a collection of
     * smaller Elements contained within their "parent" part. The
     * subclasses of Element have a conceptual "size order", and a
     * tree structure links Elements to their parents and to their
     * contained sub-images. Pages are the biggest, Regions are
     * contained within Pages, Lines are contained within Regions or
     * Pages, Words are contained within Lines, Regions, or Pages, and
     * Glyphs are contained within Words, Lines, Regions or Pages.
     *
     * Implementation subtleties:
     * --------------------------
     * Since Elements are part of a part-whole decomposition into
     * smaller and smaller images, the manner in which this
     * decomposition is implemented can have significant time and
     * space efficiency ramifications.
     *
     * One strategy would be to have each Element subclass maintain a
     * local copy of that portion of the overall image to which it
     * applies. Fields 'height' and 'width' would establish the pixel
     * dimensions, and a 'data' field could store the actual pixel
     * information. For example, each Glyph could maintain a copy of
     * that portion of the Page represented by the Glyph.
     *
     * The problem with the above naive implementation is that it
     * incurs more and more memory the more sub-divided an Element
     * becomes. Although it is common for Pages to contain just a
     * collection of Glyphs, it is also possible for a Page to contain
     * Regions that contain Lines that contain Words that contain
     * Glyphs (which might contain other Glyphs). In such a situation,
     * the pixel data representing an individual Glyph would be copied
     * (and maintained in memory) up to 5 times (stored in a Glyph,
     * stored in a Word, stored in a Line, stored in a Region, and
     * stored in the Page.
     *
     * The above memory impact can be avoided by taking a different
     * approach. Instead of each Element maintaining a separate copy
     * of its portion of an image, we note that the largest possible
     * Element is a Page, and all other subclasses of Page are
     * conceptually contained within a Page. If each Element has
     * access to its containing Page, then we could replaced the
     * 'data' field with a <x,y> coordinate identifying the topleft
     * pixel within the Page at which the sub-image begins. In
     * addition, we could maintain either a <width,height> pair, or
     * another <x,y> pair representing the bottom right corner (either
     * one will allow computation of the other). By maintaining
     * topleft relative to the Page, we do not need to maintain
     * multiple copies of the data (which is what took up most of the
     * memory in the first variant above).
     *
     * Because this second strategy significantly reduces memory
     * costs, it is the strategy currently used. However, the
     * interface is separate from the implementation, and we can
     * change the implementation as desired if it becomes necessary.
     *
     * An Element has the following fields:
     *   parent       --> the parent Element
     *   parts        --> the list of sub parts
     *   topleft      --> the <x,y> pixel coordinate within the Page image
     *                    to which this image belongs of the top-left corner
     *                    of this sub image.  <x,y> is <0,0> for Page
     *                    instances.
     *   bottomright  --> the <x,y> pixel coordinate within the Page image
     *                    to which this image belongs of the bottom-right
     *                    corner of this sub image.
     *
     * An Element can establish its Page by traversing its 'parent'
     * links until one is found that has no parent (Page instances
     * have a NULL parent).
     *
     *///******************************************************************

    class Element : public Root {
      public:
        // Types
        typedef std::list<Element*> ElementList;
        
        /**
         * Each Element specifies its parent.
         *
         * Checks are performed to ensure that the parent specified
         * can be a parent of the image in question. See the 'type'
         * method and 'parent' field for more details.
         *
         * It is possible for parent to be NULL. When the reliance on
         * C code is removed, this should only occur for Page
         * instances (which do not have an Element parent), but
         * currently the C code creates by-value Glyphs using a
         * default constructor, so we must allow them to have NULL
         * parents as well. Note that the Page constructor always
         * passes a NULL pointer to its parent constructor, and that
         * Glyph has a default value of NULL, while the other
         * subclasses require a real image.
         *
         */
        Element(Element* parent, const Coord& topleft, const Coord& size);
        virtual ~Element();

        /**
         * Marks the element as invisible - it can be ignored by all
         * processing that is directed to producing text.
         *
         */
        inline void markInvisible(bool invisible = true) { this->invisibleIs(invisible); }
        
        /**
         * Returns an integer establishing how "small" this image type
         * is, relative to other image types. It has nothing to do
         * with width or height, but instead with the conceptual size
         * of the type itself. All instances of a particular subtype
         * will always return the same value. The code is designed so
         * that Glyph returns a larger number than Word, which is
         * larger than Line, which is larger than Region, which is
         * larger than Page. This allows us to perform some sanity
         * checks on hierarchial decompositions to ensure that we
         * don't make silly structures in which Lines have Glyphs as
         * parents, etc.
         *
         * FUTURE FIX: This method should be pure-virtual, but making
         * it pure-virtual causes compilation failure (pure virtual
         * method invoked in constructor).
         *
         */
         virtual int type() const;
        
         /**
          * Add an Element to the set of parts within myself.
          *
          */
         void registerElement(Element* part);
         Element* firstElement() const { return ( this->parts().empty() ) ? NULL : this->parts().front(); }
         Element* lastElement() const  { return ( this->parts().empty() ) ? NULL : this->parts().back();  }
         
         Coord size() const { return Coord( this->width(), this->height() ); }

         /**
          * Accumulate all glyphs in this Element.
          *
          * Adds to the 'glyphs' vector all Glyph images within this
          * Element, found by recursively descending the 'parts'
          * field.
          *
          */
         void allGlyphs(std::vector<Glyph*>& glyphs);
         void allGlyphs(std::vector<const Glyph*>& glyphs) const;
        
        /**
         * Returns the Page within which this Element resides.
         *
         * If the Element is a Page, returns itself.  Otherwise,
         * the Element is a sub-region of a Page (and the Page of which
         * it is a sub-region is returned).
         *
         * This method is important because the 'topleft' and
         * 'bottomright' fields of an Element describe coordinates within
         * the Image associated with Page. That is, Element subclasses other
         * than Page do not maintain their own local Images, but instead
         * maintain information establishing what sub-region of the Page
         * Image they apply to. Doing so allows us to avoid maintaining
         * relative coordinates and redundant copies of parts.
         */
        Page* containingPage() const;
        
        /**
         * The Image that this Element is a sub-region of.
         *
         * Every subclass of Element stores topleft and bottomright
         * coordinates relative to an Image stored within a Page. This
         * method returns the Image.
         *
         * Not named 'image' to avoid confusing with the accessors in
         * Page for field 'image'.
         *
         */
        virtual const Image* pageImage() const;

        /**
         * Writes a textual representation of this element to given output stream
         *
         */
        virtual void writeText(std::ostream& os) const;
        
        /**
         * Although there are good reasons for superclasses not knowing
         * about subclasses, I find the convenience of these downcasting
         * methods too useful to avoid. Redefinitions are provided in
         * subclasses for the relevant method (Page::asPage() returns
         * 'this', Glyph::asGlyph returns 'this', etc.)
         *
         */
        virtual Page*   asPage()   { return NULL; }
        virtual Region* asRegion() { return NULL; }
        virtual Line*   asLine()   { return NULL; }
        virtual Word*   asWord()   { return NULL; }
        virtual Glyph*  asGlyph()  { return NULL; }
        
        virtual const Page*   asPage()   const { return NULL; }
        virtual const Region* asRegion() const { return NULL; }
        virtual const Line*   asLine()   const { return NULL; }
        virtual const Word*   asWord()   const { return NULL; }
        virtual const Glyph*  asGlyph()  const { return NULL; }
        
        /**
         * The height in pixels of this Element.
         *
         */
        inline int height() const { return this->bottomright().y - this->topleft().y + 1; }

        /**
         * The width in pixels of this Element.
         *
         */
        inline int width()  const { return this->bottomright().x - this->topleft().x + 1; }
        
        // **********************************************************
        // Debugging methods
        // **********************************************************
        
        /**
         * Print out the containment hierarchy for this image.
         *
         */
        void printStructure(std::ostream& os = std::cerr, const std::string& indent = "") const;
        
        /**
         * Print out information about this Element.
         *
         */
        virtual void printSummary(std::ostream& os = std::cerr, const std::string& indent = "", int index = -1) const;

        /**
         * Create .pgm files for every Glyph within this Element
         *
         */
        void writeGlyphs(const std::string& dir, const ImageArgs& adj) const;
        
        /**
         * Returns a unique string identifying this image among all
         * others.
         *
         * Consists of 'id', followed by integers separated by
         * underscores. Each integer identifies a position within a
         * container (Region, Line, Word, Glyph). The result is
         * designed to be useful in filenames.
         *
         * NOTE: This is not an efficient method - it is only meant to be
         * used for debugging purposes, not core functionality.
         *
         */
        virtual std::string id() const;

        /**
         * Returns the index of 'image' within this Element's list of
         * subparts.
         *
         * Returns -1 if the image is not an immediate sub-image of
         * this image.
         *
         */
        int findIndex(const Element* image) const;

        /**
        * Unit testing method.
        *
        * This static method should create instances of the class (and
        * instances of any other class necessary) and perform tests to
        * ensure that all methods within the class are working as
        * expected.
        *
        */
        static void test(int argc = 0, const char* argv[] = NULL);
        
        inline ElementList & partsRef() { return this->_parts; }
        
        // Accessors
        inline const Coord &   topleft() const { return this->_topleft; }
        inline const Coord &   bottomright() const { return this->_bottomright; }
        inline const ElementList& parts() const { return this->_parts; }
        inline const Element*    parent() const { return this->_parent; }
        inline bool  invisible() const { return this->_invisible; }
        
      protected:
        // Accessors
        inline void            topleftIs(const Coord & topleft) { this->_topleft = topleft; }
        inline Coord &         topleftRef() { return this->_topleft; }
        inline void            bottomrightIs(const Coord & bottomright) { this->_bottomright = bottomright; }
        inline Coord &         bottomrightRef() { return this->_bottomright; }
        inline void            partsIs(const ElementList & parts) { this->_parts = parts; }
        inline void            parentIs(Element* parent) { this->_parent = parent; }
        inline Element* &      parentRef() { return this->_parent; }
        inline void            invisibleIs(bool invisible) { this->_invisible = invisible; }
        
        // Methods
        
        // Fields
        static char* ClassName[6];
        
      private:
        // Accessors
        inline bool& invisibleRef() { return this->_invisible; }
        
        // Methods
        
        // *******************
        // State

        /**
         * The coordinate (in pixels) of the topleft corner of this
         * image within its containing Page (relative to the topleft
         * of the Page, with pixel indices starting at 0).
         *
         */
        Coord           _topleft;
        
        /**
         * The coordinate (in pixels) of the bottom right corner of
         * this image within its containing Page (relative to the
         * topleft of the Page, with pixel indices starting at 0).
         *
         */
        Coord           _bottomright;
        
        /**
         * The parent Element of this Element
         *
         * There is a hierarchial decomposition of parts associated
         * with this code, and this field maintains a link to the
         * parent in this decomposition. For example, each Glyph
         * belongs to a Page, but not necessarily directly. It could
         * belong to a Word that belongs to a Line that belongs to a
         * Region that belongs to a Page. Or it could belong directly
         * to a Page, or any variation in between these two.
         *
         * However, an image should never have as a parent another
         * Element that is conceptually "smaller" than it. For
         * example, a Word should not have a Glyph parent, a Line
         * should not have a Word or Glyph parent, etc. This is
         * enforced at runtime within the constructor by ensuring that
         * the 'type' of the child is greater than the 'type' of the
         * parent.
         *
         */
        Element*          _parent;

        /**
         * The collection of sub-Elements within this Element
         *
         * A Page contains Regions and/or Lines and/or Words and
         * Glyphs, a Region contains Lines and/or Words and Glyphs, a
         * Line contains Words and/or Glyphs, Words contain Glyphs,
         * and Glyphs may contain other Glyphs.
         *
         * Although compile-time type-checking could have been
         * provided by having individual subclasses define this field
         * with a specific type, that approach would have precluded us
         * from allowing a Page to contain Regions OR Lines OR Words
         * OR Glyphs (we would have had to pick one of them, or
         * introduce a superclass for some subset of them). We can,
         * however, provide run-time typechecking to ensure that we do
         * not try to add a Line as a child of a Word, etc. See the
         * 'registerElement' method for more.
         *
         * Note that although some subclasses could benefit from
         * having a vector to improve random access, it is more likely
         * that we will need to insert/delete sub-parts than it is
         * that we we need random access. We can use the list
         * iterators for traversal, and see the 'firstElement' and
         * 'lastElement' methods for more.
         *
         */
        ElementList _parts;
        
        /**
         * Boolean true if this element is to be considered invisible
         * during output.
         *
         * Note that since an invisible element is not displayed, it
         * is effectively deleted. All algorithms that process the
         * element with the aim of producing output can simply ignore
         * such elements during processing.
         *
         * Note that one can also remove an Element from its
         * containment hierarchy. Whether it is better to delete or
         * mark invisible depends on the context.
         *
         */
        bool            _invisible;
    };
};

# endif // Conjecture_Element_h

