# ifndef Env_h
# define Env_h Env_h

// C++ Dependencies
# include "Root.h"            // parent
# include <vector>            // uses
# include <map>               // uses
# include "Page.h"            // has-a  (needed here to define friends)

// Declaration
namespace Conjecture {

  class OCRModule;       // has-a
    
  /***********************************************************************
   * \class Env
   *
   * \brief The public front-end to the Conjecture set of classes
   *
   * This class acts as a front-end, maintaining all user-provided and
   * internal meta-information. A 'main' program can create an
   * instance of this class and interact entirely with it, with or
   * without involving the other classes in Conjecture.
   *
   **********************************************************************/

  class Env : public Root {
  public:

      // Constructors
    Env(int argc, const char* argv[]);
    virtual ~Env();

    /**
     * \brief Initialize fields that control variations in processing
     * algorithm.
     *
     */
    bool processArgs(int argc, const char** argv);

    /**
     * \brief Add a new Page to my list of pages
     *
     * Creates a Page instances, makes it the current page within
     * this Env, adds it to the 'pages' vector, and returns it.
     */
    Page* registerPage(const std::string& file);

    /**
     * Returns the currently "active" page.
     *
     */
    Page* currentPage() { return this->currentRef(); }

    /**
     * Perform any final post-processing mandated by config fields.
     *
     */
    void wrapup();

    /**
    * Unit testing method.
    *
    * This static method should create instances of the class (and
    * instances of any other class necessary) and perform tests to
    * ensure that all methods within the class are working as
    * expected.
    *
    */
    static void test(int argc = 0, const char* argv[] = NULL);

    // Accessors

  public:
    inline const std::string & moduleId() const     { return this->_moduleId; }
    inline const Page* current() const         	    { return this->_current;  }
    inline const int verbose() const           	    { return this->_verbose;  }
    inline const std::string & infile() const  	    { return this->_infile;   }
    inline const std::string & outfile() const 	    { return this->_outfile;  }
    inline const std::vector<Page*> & pages() const { return this->_pages;    }
    inline const std::string & rawdata() const 	    { return this->_rawdata;  }
    inline const char** argv() const           	    { return this->_argv;     }
    inline unsigned argc() const               	    { return this->_argc;     }

    /***
     * Returns an object that provides access to algorithm
     * implementations.
     *
     */
    inline const OCRModule* module() const { return this->_module; }

    /**
     * Factory Method for creating an appropriate Module
     *
     * Note that the default implementation is auto-generated by
     * 'ocrgen' based on the 'config/Conjecture.modules' file.
     * Subclassing should really not be necessary (and thus admittedly
     * it isn't much of a Factory Method :-).
     *
     */
    virtual OCRModule* makeModule();

  protected:
    inline void currentIs(Page* current)                  { this->_current = current;   }
    inline void moduleIdIs(const std::string & moduleId)  { this->_moduleId = moduleId; }
    inline void pagesIs(const std::vector<Page*> & pages) { this->_pages = pages;       }
    inline void verboseIs(int verbose)                    { this->_verbose = verbose;   }
    inline void outfileIs(const std::string & outfile)    { this->_outfile = outfile;   }
    inline void infileIs(const std::string & infile)      { this->_infile = infile;     }
    inline void rawdataIs(const std::string & rawdata)    { this->_rawdata = rawdata;   }
    inline void argcIs(int argc)                          { this->_argc = argc;         }
    inline void argvIs(const char** argv)                 { this->_argv = argv;         }

  private: 
    inline Page*&                currentRef()  { return this->_current;  }
    inline std::string &         moduleIdRef() { return this->_moduleId; }
    inline std::string &         rawdataRef()  { return this->_rawdata;  }
    inline int&                  verboseRef()  { return this->_verbose;  }
    inline std::string &         outfileRef()  { return this->_outfile;  }
    inline std::string &         infileRef()   { return this->_infile;   }
    inline std::vector<Page*>&   pagesRef()    { return this->_pages;    }

    /**
     * This field is interpreted in an algorithm-specific fasion.
     *
     * Initialied from the -Z command-line flag, it is primarily
     * designed for use in conjunction with Algsets that invoke
     * external programs (which themselves take command line
     * arguments to customize behavior).
     *
     * An example usage is with the current 'ocrad' implementation,
     * which invokes the external 'ocrad' program and passes this
     * field as additional argument.
     *
     */
    std::string _rawdata;

    /**
     * The name of the OCR Module to use (the value associated with
     * the -A flag).
     *
     */
    std::string     _moduleId;

    /**
     * An instance of some subclass of OCRModule, it establishes the set
     * of algorithms to use.
     *
     * The OCRModule interface defines one Factory Method for each
     * algorithm in the Conjecture framework. The 'module' method
     * creates an instance of a specific subclass in the appropriate
     * Strategy hierarchy. Every algorithm has an associated Strategy
     * hierarchy.
     *
     */
    OCRModule* _module;
    inline void moduleIs(OCRModule* module) { this->_module = module; }
    inline OCRModule* & moduleRef() { return this->_module; }

    /**
     * The file to write output to.  Specified via the -o flag.
     *
     */
    std::string _outfile;

    /**
     * The file to read input form.  Specified via the -i flag (or first arg after flags)
     *
     */
    std::string _infile;

    /**
     * Current page being processed.
     *
     */
    Page* _current; 

    /**
     * List of all pages in environment.  
     *
     * Often, there is only one page, but support exists for
     * for dealing with multiple pages.
     *
     */
    std::vector<Page*> _pages;   

    /**
     * The level of verbosity specified for this execution.
     *
     * A value of 0 means no diagnostic output is printed.
     * A value of 1 means major diagnostic items are noted.
     * Increasing values provide more and more detailed diagnostics.
     *
     */
    int _verbose;

    /**
     * The number of command-line arguments
     *
     */
    unsigned _argc;

    /**
     * The raw array of command-line arguments.
     *
     * This is stored in the Env class because some Modules may want
     * to process command-line flags themselves. Note that if a
     * particular third-party OCR uses command-line arguments, they
     * will almost certainly conflict with command-line arguments in
     * Env, so this field is problematic and may be removed.
     *
     */
    const char** _argv;

  };
};

// C++ Dependencies
# include "Page.h" // has-a

# endif // Env_h

