# ifndef Glyph_h
# define Glyph_h Glyph_h

// C Dependencies
//  - FUTURE FIX: these headers are only needed while the C++ interface
//    relies on the underlying C code.

# include "pnm.h"
# include "gocr.h"  // Box

// C++ Dependencies
# include <vector>
# include "Element.h"
# include "ImageArgs.h"
# include "GlyphFeatures.h"  // LineType, has-a

namespace Conjecture {

    /**
     * Experimental macro - not sure if it will stay though.
     *   The passed-in 'self' variable must be of type 'Element*' (or subclass, naturally)
     *   The passed-in 'glyph' variable does NOT need to be declared - it
     *     is declared internally (as a 'const Glyph *').
     *
     */
# define GlyphLoopStart(self, glyph) \
{ \
    std::vector<Glyph*> __glyphs; \
    (self)->allGlyphs(__glyphs); \
    for (std::vector<Glyph*>::iterator __it = __glyphs.begin(); __it != __glyphs.end(); ++__it) { \
      Glyph* glyph = *__it;

# define ConstGlyphLoopStart(self, glyph) \
{ \
    std::vector<const Glyph*> __glyphs; \
    (self)->allGlyphs(__glyphs); \
    for (std::vector<const Glyph*>::const_iterator __it = __glyphs.begin(); __it != __glyphs.end(); ++__it) { \
      const Glyph* glyph = *__it;

# define GlyphLoopEnd \
    } \
}

  /*********************************************************************
  * \class Glyph
  *
  * \brief A rectangular region of the overall image that has been
  *  identified as representing a single character.
  *
  * The Box class represents all information associated with the
  * graphical reprsentation of a to-be-identified character.  The
  * correctness of this box (i.e. whether it really does represent
  * a single character) depends on:
  *  - the accuracy of line detection algorithms (currently pretty good)
  *  - the accuracy of box detecting algorithms (currently ok,
  *    improvements possible)
  *
  * Ths class is responsible for converting an image of a character
  * into a unicode character value, and as such is at the heart of the
  * OCR code.
  *///******************************************************************

  class Glyph : public Element {
  public:
    // ********
    // Types
    class Char {
      public:
      Char(wchar_t c,int w) { this->c = c; this->weight = w; }
      wchar_t c;
      int weight;
    };

    // *******
    // Methods

    /**
     * Parent allowed to be NULL only until dependence on C code is
     * removed. The 'internal' param is also temporary (it is a Box
     * struct if provided).
     *
     */
    Glyph(Element* parent = NULL, void* internal = NULL);

    /**
     * Non-standard copy constructor (may switch to pass-by-ref later)
     *  - arg intentionally non-const due to parent-link issues
     *
     */
    Glyph(Glyph* glyph);

    /**
     * Allows Image instances to be converted to Glyph instances
     * safely without requiring explicit down-casting.
     *
     */
    virtual Glyph* asGlyph()             { return this; }
    virtual const Glyph* asGlyph() const { return this; }

    inline void setUnicode(wchar_t unicode) { this->unicodeIs(unicode); }

    virtual void printSummary(std::ostream& os = std::cerr, const std::string& indent = "", int index = -1) const;

    void addChar(const Glyph::Char& chr);

    /**
     * Returns a categorization of this glyph within its line.
     *
     * See the LineType documentation for semantics.
     *
     */
    GlyphFeatures::LineType lineType() const;

    u2 topY() const { return this->topleft().y; }
    u2 bottomY() const { return this->bottomright().y; }
    u2 leftX() const { return this->topleft().x; }
    u2 rightX() const { return this->bottomright().x; }

    // ---------------------------------------------------------------
    // DEBUGGING METHODS

    /**
     * Write Glyph image to disk.
     *
     * Creates a file (with prefix 'filebase', suffix '.pgm') for the
     * sub-region of the pageImage representing this Glyph. The output
     * may be magnified, thresholded, and/or filtered. See the
     * Image::Image constructor documentation for more.
     *
     */
    void writeImage(const std::string& filebase, const ImageArgs& config);

    /**
     * Writes a textual representation of this element to given output
     * stream.
     *
     */
    virtual void writeText(std::ostream& os) const;

    /**
     * Unit testing method.
     *
     * This static method should create instances of the class (and
     * instances of any other class necessary) and perform tests to
     * ensure that all methods within the class are working as
     * expected.
     *
     */
    static void test(int argc = 0, const char* argv[] = NULL);

    wchar_t unicode() const { return this->_unicode; }
    const std::vector<Char>& charset() const { return this->_charset; }

    // Accessors
    inline int  best() const { return this->_best; }
    virtual int type() const { return Type; }
    // we provide 'box' publically because it is temporary - need to
    // remove it anyway, and Page often needs the Box within Glyph.
    inline const Box* box() const { return this->_box; }
    inline void boxIs(Box* box)   { this->_box = box; }
    inline Box* & boxRef()        { return this->_box; }

  protected:
    // ********
    // Methods

    // ********
    // Accessors
    inline void bestIs(const int & best) { this->_best = best; }
    inline int &           bestRef() { return this->_best; }
    wchar_t& unicodeRef() { return this->_unicode; }
    void unicodeIs(wchar_t unicode) { this->_unicode = unicode; }
    std::vector<Char>& charset() { return this->_charset; }

  private:
    // ********
    // Methods

    // ********
    // Accessors

    // ********
    // State
    wchar_t           _unicode;

    /**
     * The Type field allows Element to enforce some containment
     * restrictions on the Element has-a hierarchy
     *
     * For exampel, a Page can contain a Region, Line, Word or Glyph,
     * but a Glyph cannot contain a Word or Line or Region or Page.
     * Both Page and Word inherit the same 'parts' field (a list of
     * Element*), so we must enforce the above restrictions at
     * runtime. Element only allows instances to be added to 'parts'
     * if its 'type' value is greater than the 'type' value of the
     * containing Element.
     *
     */
    static const ElementType Type = ELEMENT_GLYPH;

    // Are any of these really necessary?
    int                _best; // inex into _charset (best match)
    Box*              _box;   // old gocr data-structure
    std::vector<Char> _charset;
  };
};

# endif // Glyph_h

