
// C++ Dependencies
# include <iostream>
# include <sstream>
# include "Image.h"  
# include "Element.h"  // Coord
# include "Env.h"      // FUTURE FIX: only needed temporarily for 'test'
# include "System.h"   // Mkpath

// C Dependencies (temporary)
# include "pnm.h"      // readpnm, writepgm, writeppm, writepbm
# include "pgm2asc.h"  // copybox


using namespace std;

// ******************************************************************
// CONSTRUCTORS
namespace Conjecture {
  Image::Image(const string& file, bool verbose) {
    if ( verbose ) { cerr << "NOTE: Reading PNM file " << file << endl; }
    char* fn  = const_cast<char*>(file.c_str());
    pix* data = (pix*)malloc(sizeof(pix));
    this->dataIs(data);
    readpgm(fn, data, 0);
  }

  Image::Image(const Image* src, const Coord& topleft, const Coord& bottomright, const ImageArgs& config) {
    // FUTURE FIX: Currently relying on gocr C data-structures, but no
    // longer relying on C functions to implement this. Must remove
    // reliance on C data-structures asap.
    uint8_t magnification = config.magnification();

    bool verbose = false; // plan for future.  Should we add verbose to ImageArgs?
    if ( verbose ) { 
      Coord size(src->data()->x,src->data()->y);
      cerr << "NOTE: Creating sub-image from " << topleft.str() 
           << " to " << bottomright.str() 
           << endl 
           << "NOTE:    from within image of size " << size.str();
      if ( magnification > 1 ) {
        cerr << " (magnification " << magnification << ")";
      }
      cerr << endl;
    }

    if ( src->bytesPerPixel() != 1 ) {
	// Not yet handling color - does the code below work for color
	// as long as magnitude is 1, or does it need to be changed to
	// support basic copying too?
	cerr << "NOTE: bytesPerPixel = " << src->bytesPerPixel() << endl;
	throw string("UnsupportedFeatureException: Not currently handling the copying and/or magnification of color images");

    } else {
      // The current implementation uses a 'pixmap' data-structure
      // nested within this Image class.  Although this code does not
      // rely on the low-level gocr C functions, we maintain the
      // pix structure because other methods in this class do still rely
      // on it.  Note that we are managing to start separating the code
      // though.
      
      // Obtain the 'pixmap' structure to be initialized (from 'this')
      pix* subdata = (pix*) malloc(sizeof(pix));  // Must use malloc because gocr does - be careful about unintentionally mixing new and malloc - nightmares!
      this->dataIs(subdata);
      
      // Prepare the fields in 'subdata'.  
      int dx = bottomright.x - topleft.x + 1;
      int dy = bottomright.y - topleft.y + 1;

      // Magnification multiplies both dimensions.
      subdata->x = dx * magnification;
      subdata->y = dy * magnification;

      int bytes = subdata->x*subdata->y;
      subdata->p = (unsigned char*)malloc(bytes);

      // Now copy from source into destination. Note that we are not
      // using the low-level copybox, nor are we removing low bits or
      // doing pixel filtering.  Such processing will be done by
      // other methods.
      int x0 = topleft.x;
      int y0 = topleft.y;
      for (int y = 0; y < dy; y++) {
        for (int x = 0; x < dx; x++) {
          int val = config.apply(src, Coord(x+x0,y+y0));
          for (int ym = 0; ym < magnification; ++ym) {
            for (int xm = 0; xm < magnification; ++xm) {
              // DEBUG: cerr << Coord(x*magnification+xm,y*magnification+ym).str() << " from " << Coord(x+x0,y+y0).str() << endl;
              this->set(x*magnification+xm,y*magnification+ym, val);
            }
          }
        }
      }
    }
  }

  Image::~Image() {
    // The temporary 'data' field is by-value, but its 'p' field
    // is dynamically allocated via 'malloc'.
    if (this->data()->p) {
      free(this->data()->p);
      this->dataRef()->p = NULL;
    }
  }
  
  // ******************************************************************
  // ACCESSORS
  
  
  // ******************************************************************
  // METHODS

  bool Image::writeGrey(const string& filebase, bool verbose) const {
      bool result = false;
      // Establish the file
      string file = filebase + ".pgm";
      // Ensure its directory exists.
      if ( System::Mkpath(file) ) {
	  if ( verbose ) { cerr << "NOTE: Writing PGM (grey)  file " << file << endl; }
	  pix* data = const_cast<pix*>(this->data());
	  // FUTURE FIX: This type violation is forced by gocr's bad signature for 'writepgm'
	  char* fn  = const_cast<char*>(file.c_str()); 
	  result = writepgm(fn,data);
      } else {
	  result = false;
      }
      return result;
  }

  bool Image::writeColor(const string& filebase, bool verbose) const {
      bool result = false;
      // Establish the file
      string file = filebase + ".ppm";
      // Ensure its directory exists.
      if ( System::Mkpath(file) ) {
	  if ( verbose ) { cerr << "NOTE: Writing PPM (grey)  file " << file << endl; }
	  pix* data = const_cast<pix*>(this->data());
	  // FUTURE FIX: This type violation is forced by gocr's bad signature for 'writeppm'
	  char* fn  = const_cast<char*>(file.c_str()); 
	  result = writeppm(fn,data);
      } else {
	  result = false;
      }
      return result;
  }

  bool Image::writeMono(const string& filebase, bool verbose) const {
      bool result = false;
      // Establish the file
      string file = filebase + ".pbm";
      // Ensure its directory exists.
      if ( System::Mkpath(file) ) {
	  if ( verbose ) { cerr << "NOTE: Writing PBM (grey)  file " << file << endl; }
	  pix* data = const_cast<pix*>(this->data());
	  // FUTURE FIX: This type violation is forced by gocr's bad signature for 'writepbm'
	  char* fn  = const_cast<char*>(file.c_str()); 
	  result = writepbm(fn,data);
      } else {
	  result = false;
      }
      return result;
  }

  void Image::test(int argc, const char* argv[]) {
    bool verbose = true;
    Image* im = new Image("test.pnm", verbose);

    Coord tl = Coord(0,0);
    Coord br = im->size();

    Image subim(im, tl, br, ImageArgs(1));
    subim.writeGrey("tmp/outsub", verbose);

    for (int thresh = 100; thresh < 255; thresh += 5) {
	Image subim(im, tl, br, ImageArgs(2,thresh));
	stringstream s;
	s << "tmp/outsub-" << thresh;
	subim.writeGrey(s.str(), verbose);
    }

    Image subim3(im, tl, Coord(100,100), ImageArgs(3,0,15));
    subim3.writeGrey("tmp/outsub3", verbose);

    im->writeGrey("tmp/out", verbose);
    im->writeMono("tmp/out", verbose);
    //im->writeColor("tmp/out", verbose);

    delete im;
  }



};

