# ifndef ImageArgs_h
# define ImageArgs_h ImageArgs_h

# include "Root.h"
# include "Coord.h"  // uint8_t, uint16_t

namespace Conjecture {
    class Image;
    /**********************************************************************
     * \class ImageArgs
     *
     * \brief A tiny service class that stores image and
     * pixel-filtering information.
     *
     * Instances of this class are often used as literal arguments to
     * Image-related methods.  See Image::Image and Glyph::writeImage
     * for examples.
     *
     * This class may be removed when a more general Image
     * infrastructure is in place!
     *
     ********************************************************************/

    // FUTURE FIX: Should the max values of 'threshold' and 'quantize'
    // be downgraded to 255? Currently uint16_t values to support max
    // intensities up to 65535, but this is probably entirely
    // unnecessary. If we put them down to uint8_t's, the structure as-is
    // would fit in a word.  But more fields are likely to be added.
    
    // FUTURE FIX: Should we add a 'verbose' field to this class?
    // Would help in Image::Image.

    class ImageArgs : public Root {
        
      public:
        /**
         * Constructor (in default mode, provides an ImageArgs that does
         * no pixel modification).
         *
         */
        ImageArgs(uint8_t magnification = 1, uint16_t threshold = 0, uint16_t quantize = 0, uint8_t filter = 0, uint8_t dustsize = 0);
        
        uint8_t  magnification() const   { return this->_magnification; }
        uint16_t threshold()     const   { return this->_threshold;     }
        uint16_t quantize()      const   { return this->_quantize;      }
        uint8_t  filter()        const   { return this->_filter;        }
        uint8_t  dustsize()      const   { return this->_dustsize;      }
        
        uint8_t&  magnification()        { return this->_magnification; }
        uint16_t& threshold()            { return this->_threshold;     }
        uint16_t& quantize()             { return this->_quantize;      }
        uint8_t&  filter()               { return this->_filter;        }
        uint8_t&  dustsize()             { return this->_dustsize;      }
        
        /**
         * Given an image and position, returns a new color value for
         * that pixel position based on the values of fields.
         *
         * Note that the order in which the operations is performed
         * significantly affects results.  The order is always:
         * filter, quantize, threshold, dustsize.
         *
         */
        uint16_t apply(const Image* image, const Coord& pos) const;

        /**
         * FUTURE FIX: It is possible for a particular image to have a a
         * non-standard maximum intensity, so specifying MaxVal here may
         * be problematic. Note that it totally ignores the idea of 2-byte
         * color values, for example.  Do we pass maxval as an optional
         * 2nd parameter to 'adjust'???
         *
         */
        static const uint16_t MaxVal;
        static const uint16_t MinVal;

      private:
        /**
         * A value greater than one indicates how many times each pixel
         * should be replicated in both x and y directions. Since
         * magnification is meant to provide an easier-to-see view of the
         * exact pixel values, there is no aliasing or smoothing or any
         * sophisticated pixel modification performed like there would be
         * if it was a general-purpose resize mechanism.
         *
         */
        uint8_t _magnification;

        /**
         * Non-zero indicates that pixel values greater-equal threshold
         * should be given maximum intensity (white), and values less
         * than threshold should be given minimum intensity (black).
         *
         */
        uint16_t _threshold;

        /**
         * Non-zero indicates that pixel values should be logically anded
         * with the negation of this value. 'quantize' is almost always
         * 2^k-1 for some small k, but an arbitrary value is allowed
         * (although most masks will produce weird and unintuitive
         * effects).
         *
         */
        uint16_t _quantize;

        /**
         * If non-zero, this indicates that pixel filtering should be
         * applied to the image. The value of filter dictates what kind
         * of filter to apply:
         *   filter == 1   --> 3x3 filter (see gocr code) - FUTURE FIX: Must add this filter code to Image soon!
         *   filter == 2   --> 5x5 filter (to be explored)
         *
         */
        uint8_t _filter;
        
        /**
         * If non-zero, this indicates ... Experimental field - may be
         * removed
         *
         */
        uint8_t _dustsize;

        /**
         * A string representation of the configuration values
         *
         */
        std::string str() const; // printable representation
    };
};

# endif // ImageArgs_h

