# ifndef Line_h
# define Line_h Line_h

// Dependencies
#include "Element.h" // parent

// Declaration
namespace Conjecture {

    /*******************************************************************
     * \class Line
     *
     * \brief A sub-region of a Page (or Region) containing
     * Glyphs or Words containing Glyphs.
     *
     * NOTE: Whether a Page is segmented into Glyphs first, then into
     * Lines, or Lines first, then Glyphs, depends on the particular
     * segmentation/identification/formatting algorithms used.
     *
     *******************************************************************/

    class Line : public Element {
      public:
        // Constructors/Destructors
        Line(Element* parent, const Coord& topleft, const Coord& bottomright, u2 bottom, u2 base, u2 mid, u2 top);
        
        // Accessors
        inline const u2 &      topY() const { return this->_topY; }
        inline const u2 &      midY() const { return this->_midY; }
        inline const u2 &      baseY() const { return this->_baseY; }
        inline const u2 &      bottomY() const { return this->_bottomY; }
        virtual int type() const { return Type; }

        /**
         * Allows Image instances to be converted to Line instances
         * safely without requiring explicit down-casting.
         *
         */
        virtual Line*   asLine()   { return this; }
        virtual const Line*   asLine() const { return this; }

        /**
         * Print summary of this line
         *
         * Normally recurses by summarizing Glyphs within it.  Specify an
         * index of -2 (or less) to avoid this recursion.
         *
         */
        virtual void printSummary(std::ostream& os = std::cerr, const std::string& indent = "", int index = -1) const;

        /**
         * Unit testing method.
         *
         * This static method should create instances of the class (and
         * instances of any other class necessary) and perform tests to
         * ensure that all methods within the class are working as
         * expected.
         *
         */
        static void test(int argc = 0, const char* argv[] = NULL);

        /**
         * Writes a textual representation of this element to given output stream
         *
         */
        virtual void writeText(std::ostream& os) const;
        
      protected:
        // Accessors
        inline void            topYIs(const u2 & topY) { this->_topY = topY; }
        inline void            midYIs(const u2 & midY) { this->_midY = midY; }
        inline void            baseYIs(const u2 & baseY) { this->_baseY = baseY; }
        inline void            bottomYIs(const u2 & bottomY) { this->_bottomY = bottomY; }
        
        // Methods

      private:
        // Accessors
        inline u2 &            topYRef() { return this->_topY; }
        inline u2 &            midYRef() { return this->_midY; }
        inline u2 &            baseYRef() { return this->_baseY; }
        inline u2 &            bottomYRef() { return this->_bottomY; }
        
        // Methods
        
        // State

        /**
         * The vertical position (relative to the top of the image) of
         * the top of this line (the highest portion of letters like
         * 'l', 't', 'A', etc.)
         *
         */
        u2              _topY;

        /**
         * The vertical position (relative to the top of the image) of
         * the mid-line (the vertical position of the highest portion
         * of letters like 'a', 'm' and 'x').
         *
         */
        u2              _midY;

        /**
         * The vertical position (relative to the top of the image) of
         * the base-line (the vertical position on which letters like
         * 'a','m' and 'x' lie.
         *       
         */
        u2              _baseY;

        /**
         * The vertical position (relative to the top of the image) of
         * the lowest portion of this line (the portion below letters
         * like 'g', 'p', 'q' and 'y'.
         *
         */
        u2              _bottomY;
        static const ElementType Type = ELEMENT_LINE;
    };
};

# endif // Line_h

