# include <iomanip>   // setw
# include <fstream>   // file writing

# include "Page.h"    // me
# include "Line.h"
# include "Word.h"
# include "Glyph.h"
# include "Image.h"

# include "OCRModule.h"

using namespace std;

namespace Conjecture {

    Page::Page(Env* env, const string& file) : Element(NULL, Coord(0,0), Coord(0,0)) {
        // The Coord(0,0) init-list args are dirty, but the fields ARE
        // properly initialized within this constructor!.
        
        // Every Page maintains its environment
        this->envIs(env);
        
        // Initialize a Page instance from the given image file.
        this->fileIs(file);

        // We create an Image instance from 'file'
        bool verbose = env->verbose();
        Image* image = new Image(file,verbose);
        this->imageIs(image);
        
        // Since the 'topleft' and 'bottomright' fields are inherited
        // from Element, there is no reason not to use them, even
        // though for the Page class they aren't strictly necessary,
        // as we can obtain them from the underlying Image.
        this->topleftIs(Coord(0,0));
        this->bottomrightIs(image->size()-1);
    }
    
    Page::Page(Env* env, const string& imagefile, const string& validfile) : Element(NULL,Coord(0,0),Coord(0,0)) {
        // This special constructor performs something quite different
        // than normal - it creates an entire containment hierarchy of
        // Lines, Words and Glyphs (with already-assigned unicode
        // values).  It can be used as a "training" set, representing
        // the structure and unicode assignment that an OCR *should*
        // produce for a particular input image if that OCR were
        // 100% accurate.
        
        // Every Page maintains its environment
        this->envIs(env);
        
        // Store the input image file
        this->fileIs(imagefile);

        // Now initialize this Page into Lines of Words of Glyphs
        // based on the validfile contents.
        Coord fake(0,0);

        ifstream is(validfile.c_str());
        cerr << "NOTE: Reading '" << validfile << "'" << endl;
        if (!is) {
            throw string("InvalidFileException: Unknown file ")+validfile;
        } else {
            // Current line
            Line* line = new Line(this, fake, fake, 0,0,0,0);

            // This is a problem if 'validfile' is empty (adds a line
            // when we shouldn't.
            this->registerElement(line);

            // Current Word
            Word* word = new Word(this,fake,fake);
            line->registerElement(word);

            while (! is.eof()) {
                int cv = is.get();
                if ( cv < 0 ) { break; }
                char c = (char)cv;

                if ( c == ' ' ) {
                    // space means a new Word. For now, multiple
                    // adjacent spaces produce multiple Word
                    // instances, but we may provide a different
                    // solution later.
                    word = new Word(this,fake,fake);
                    line->registerElement(word);

                } else if ( c == '\n' ) {
                    // newline means new Line :-)
                    line = new Line(this, fake,fake, 0,0,0,0);
                    this->registerElement(line);
                    word = new Word(line, fake,fake);
                    line->registerElement(word);

                } else if ( c == '\t' ) {
                    // Tabs are erroneous
                    cerr << "ERROR: Encountered a tab in " << validfile << " which is disallowed" << endl;
                    exit(1);

                } else {
                    // Normal character - add a glyph
                    Glyph* glyph = new Glyph(word);
                    glyph->setUnicode(c);
                    word->registerElement(glyph);

                }
            }
        }
    }

    void Page::writeText(ostream& os) const {
        // This naive implementation doesn't usually to the proper
        // thing - subclasses reimplement as necessary.
        for (ElementList::const_iterator it = this->parts().begin(); it != this->parts().end(); ++it) {
            (*it)->writeText(os);
        }
    }

    // Do everything to convert this page into text.
    bool Page::process() {
        return this->algorithms()->process()->execute(this);
    }
    
    // Segment Page into Glyphs
    bool Page::segment() {
        return this->algorithms()->segment()->execute(this);
    }
    
    // Map Glyphs to unicode
    bool Page::identify() {
        return this->algorithms()->identify()->execute(this);
    }
    
    // Format unicode into lines and words
    bool Page::format() {
        return this->algorithms()->format()->execute(this);
    }
    
    void Page::printSummary(ostream& os, const string& indent, int index) const {
        // For now, we simply call the default implementation in
        // Element, which recurses into subparts until Glyphs are
        // reached, which defines a more specific 'printSummary'
        // method.
        this->Element::printSummary(os,indent,index);
        
        // FUTURE FIX: Add more code here to print out page-specific
        // information before or after printing out the containment
        // structure.
    }
    
    OCRModule* Page::algorithms() const {
        // FUTURE FIX: Fix this const_cast - friendship or provide public
        // non-const access to factory in Env.
        return const_cast<OCRModule*>(this->env()->module());
    }
    
    void Page::test(int argc, const char* argv[]) {
        // Advance past <exec> -T (and pretend the -T arg is "exec"
        argc += 2;  argv = &(argv[2]);
        
        if ( argc > 1 ) {
            // Temporary - testing the new Page constructor
            Page p(NULL, argv[1], argv[2]);
            p.writeText(cerr);
            
        } else {
            // Hard-coded test that requires 'test.pnm' to exist.
            // Future fix - make this less fragile!
            static const int locargc = 5;
            static const char* locargv[locargc] = { "exec", "-i", "test.pnm", "-o", "test.ocr" };
            Env * env = new Env(locargc,locargv);
            string fname = env->infile();
            cerr << "NOTE: Loading " << fname << endl;
            Page* page = env->currentPage();
            page->segment();
            
            page->writeGlyphs("tmp", 4);
        }
    }
  
};
