using namespace std;
# include <iostream>
# include <sstream>  
# include "BitOverlay.h"
# include <math.h>   // pow (for testing)
# include <stdlib.h> // rand (for testing)

// ####################################################################
// # Conjecture: An Extensible Optical Character Recognition Framework  #
// #                                                                  #
// # Copyright: Lesser GNU Public License (LGPL)                      #
// #                                                                  #
// # Overview:                                                        #
// #  - Conjecture is a collection of C++ classes, and an OCR engine    #
// #  - Conjecture is designed to allow customization at all levels     #
// #    and to encourage individuals to contribute incremental        #
// #    improvements in algorithms.                                   #
// #  - Overall design discussions can be found in                    #
// #      $SNROOT/docs/doxygen/html/index.html                        #
// #  - Conjecture is designed to interact with and build on other      #
// #    open-source OCR programs.                                     #
// ####################################################################

// Include
# include "CpuTimer.h"

namespace Conjecture {

    // IMPORTANT: This must be defined as a macro, not a C++ function,
    // in order to get accurate timing results.  Using a function
    // requires pointers to functions, which both precludes inlining,
    // and swamps timing with the function dereference.
    # define CompareEfficiency(func1, func2, iters, timer, tmp, speedup) { \
	cerr << "NOTE:   Timing first function" << endl; \
	timer.start(); for (unsigned i = 0; i < iters; ++i) { tmp = (*func1)(i); }; timer.stop(); unsigned time1 = timer.microtime(); \
	cerr << "NOTE:   Timing second function" << endl; \
	timer.start(); for (unsigned i = 0; i < iters; ++i) { tmp = (*func2)(i); }; timer.stop(); unsigned time2 = timer.microtime(); \
	speedup = ((float)time2/(float)time1); \
    }

    // Interface
    BitOverlay::BitPos BitOverlay::FindFirstHigh(Layout val) {
	// Binary search implementation of FFS.  Presumably much
	// faster than:
	//   result = int(log2(val)+0.00001);
	// VERIFY THIS! 

	int result = npos;
	// Having to incur an extra test just to test for no bits
	// is unfortunate!  Can we assume that 'val' will never be
	// 0 so we don't have to make this test?? Probably not.
	if ( val ) {
	    // Val is non-zero
	    if ( val & 0xffff0000 ) {
		// The first bit is in the high-order 16-bits
	    	if ( val & 0xff000000 ) {
		    // The first bit is in the high-order 8-bits
		    if ( val & 0xf0000000 ) {
			// The first bit is in the high-order 4-bits
			if ( val & 0xc0000000 ) {
			    // The first bit is in the high-order 2-bits
			    result = (val&0x80000000) ? 31 : 30;
			} else {
			    result = (val&0x20000000) ? 29 : 28;
			}
		    } else {
			if ( val & 0xc000000 ) {
			    result = (val&0x8000000) ? 27 : 26;
			} else {
			    result = (val&0x2000000) ? 25 : 24;
			}
		    }
	    	} else {
		    if ( val & 0xf00000 ) {
			if ( val & 0xc00000 ) {
			    result = (val&0x800000) ? 23 : 22;
			} else {
			    result = (val&0x200000) ? 21 : 20;
			}
		    } else {
			if ( val & 0xc ) {
			    result = (val&0x80000) ? 19 : 18;
			} else {
			    result = (val&0x20000) ? 17 : 16;
			}
		    }
	    	}   
	    } else {
		// The first bit is in the low-order 16-bits
	    	if ( val & 0xff00 ) {
		    if ( val & 0xf000 ) {
			if ( val & 0xc000 ) {
			    result = (val&0x8000) ? 15 : 14;
			} else {
			    result = (val&0x2000) ? 13 : 12;
			}
		    } else {
			if ( val & 0xc00 ) {
			    result = (val&0x800) ? 11 : 10;
			} else {
			    result = (val&0x200) ? 9 : 8;
			}
		    }
	    	} else {
		    if ( val & 0xf0 ) {
			if ( val & 0xc0 ) {
			    result = (val&0x80) ? 7 : 6;
			} else {
			    result = (val&0x20) ? 5 : 4;
			}
		    } else {
			if ( val & 0xc ) {
			    result = (val&0x8) ? 3 : 2;
			} else {
			    result = (val&0x2) ? 1 : 0;
			}
		    }
	    	}   
	    }
	}
	return result;
    }

    BitOverlay::BitPos BitOverlay::FindFirstHighSlow(Layout val) {
	BitPos result = 31;
	unsigned mask = (1<<result);
	while (mask) {
	    if ( val&mask ) { return result; }
	    --result;
	    mask = mask >> 1;
	}
	return npos;
    }

    BitOverlay::BitPos BitOverlay::FindFirstLow(Layout val) {
	// NOTE: Not using a binary search like for findFirstHigh.
	// It is suspected that the first low bit set will be
	// close to position 0.  It is, however, possible that
	// a binary search will give overall better performance,
	// especially when testing a layout containing only
	// one bit.
	BitPos result = npos;
	unsigned mask = 1;
	for (int i = 0; i < 32; ++i) {
	    if ( val & mask ) {
		result = i;
		break;
	    }
	    mask = mask << 1;
	}
	return result;
    }

    BitOverlay::BitCount BitOverlay::PopCountSlow(Layout val) {
	// Return the number of "on" pixels.
	BitCount result = 0;
	unsigned mask = (1<<31);
	while (mask) {
	    if ( val&mask ) { ++result; }
	    mask = mask >> 1;
	}
	return result;
    }

    BitOverlay::BitCount BitOverlay::Regions(Layout val, int* largest, int* all) {
	// FUTURE FIX: This is totally hacked - certainly
	// redundant code here, and probably a much more
	// efficient mechanism.
	BitCount result = 0;
	int      c      = 0;
	int      max    = 0;
	unsigned mask   = (1<<31);
	bool     active = false;
	while (mask) {
	    if ( val&mask ) {
		++c;
		active = true;
	    } else if ( active ) {
		// The current pixel is the first zero after a one
		active = false;
		if ( c > max ) {
		    max = c;
		}
		if ( all ) {
		    all[result] = c;
		}
		++result;
		c = 0;
	    }
	    mask = mask >> 1;
	}
	// If we have c non-zero it is an final region
	if ( c > 0 ) { 
	    if ( c > max ) { max = c; }
	    if ( all ) { all[result] = c; }
	    ++result;
	}
	// Record the max-sized region
	if ( largest != NULL ) {
	    *largest = max;
	}
	return result;
    }

    string BitOverlay::Str(Layout val) {
	// Produce the bit representation of 'val' (0's and 1's only).
	stringstream s;
	unsigned mask = (1<<31);
	while (mask) {
	    s << ((val&mask) ? '1' : '0');
	    mask = mask >> 1;
	}
	return s.str();
    }

    void BitOverlay::Test(int argc, const char* argv[]) {
	// Initialize random generator - useful in various tests
	srand(time(NULL));

	// Establish a timer so we can compare different implementation
	// strategies.
	CpuTimer timer;

	// *********************************************************
	// Testing FFS
	BitOverlay::BitPos pos;
	if ( (pos = FindFirstHigh(1)) != 0 ) { cerr << "ERROR: FindFirstHigh(1) = " << pos << " != 0" << endl; }
	if ( (pos = FindFirstHigh(5)) != 2 ) { cerr << "ERROR: FindFirstHigh(5) = " << pos << " != 2" << endl; }
	if ( (pos = FindFirstHigh(1392)) != 10 ) { cerr << "ERROR: FindFirstHigh(1392) = " << (int)pos << " != 10" << endl; }
	for (int i = 0; i < 100; ++i) {
	    BitOverlay::Layout val = rand();
	    // if we always use val as-is, it will always be quite large
	    if ( i < 25 ) { val >>= 23; }
	    else if ( i < 50 ) { val >>= 15; }

	    pos = FindFirstHigh(val);
	    BitOverlay::BitCount pc1 = PopCount(val);
	    BitOverlay::BitCount pc2 = PopCountSlow(val);

	    int plog = (val) ? (int)(log2(val)+0.00001) : BitOverlay::npos;
	    cerr << setw(10) << val << " = " << Str(val) << " -> ffs = " << dec << setw(2) << pos << " popc = " << pc1 << endl;
	    if ( pos != plog ) { cerr << "ERROR: mismatch between ffs() and plog = " << plog << " for " << val << endl; }

	    if ( pc1 != pc2 ) { cerr << "ERROR: val=" << val << " produces popc=" << pc1 << " popc2=" << pc2 << endl; }
	}

	// ************************************************************
	float speedup;

	cerr << "NOTE: Timing FindFirstHigh" << endl;
	CompareEfficiency(FindFirstHigh, FindFirstHighSlow, 1000000000, timer, pos, speedup);
	cerr << "NOTE:   Fast version is " << speedup << " times faster " << endl;

	BitOverlay::BitCount cnt;
	cerr << "NOTE: Timing PopC (bruno fast)" << endl;
	CompareEfficiency(PopCount, PopCountSlow, 500000000, timer, cnt, speedup);
	cerr << "NOTE:   Fast version is " << speedup << " times faster " << endl;
	
    }
};

