# ifndef BitOverlay_h
# define BitOverlay_h BitInfo_h

// # Created 2006/06/09 by wade

# include "Root.h"  // parent
# include <string>

namespace Conjecture {
    
    /********************************************************************
     * \class BitOverlay
     *
     * \brief Hyper-low-level bit analysis
     *
     * This class provides a variety of routines for analyzing the
     * bits within a 32-bit value.
     *
     * There are both static and instance versions of each method.
     *
     * An "on" bit is 1, an "off" bit is 0.
     *
     * IMPORTANT: This class can be used as simply a collection of
     * static methods, OR it can be used as an "overlay", a means of
     * adding semantics and functionality to an arbitrary integer.
     * However, in order for it to be an overlay, NO VIRTUAL METHODS!
     * We do not want a vtbl here!  Includes parents, of course.
     * Because of this, statically typing a variable as BitOverlay
     * and assigning it an instance of the subclass BitInfo is NOT
     * recommended.
     *
     * FUTURE WORK: Analyze 
     *     http://graphics.stanford.edu/~seander/bithacks.html
     * and incorporate useful methods from there into this class!
     *
     * See Also: BitOverlay64
     *
     *******************************************************************/
    class BitOverlay : public Root {
      public:
	/**
	 * The type describing the underlying value being analyzed.
	 *
	 * A typedef to make a BitOverlay64 implementation look as
	 * similar as possible to BitOverlay (and to make potential
	 * transitioning to a template implementation easier).
	 *
	 */
        typedef unsigned int Layout;

	/**
	 * A bit position.
	 *
	 * Bit position 0 is the lowest-order bit, and bit position
	 * 31 is the highest-order-bit. The special value 'npos' is
	 * used to represent "not present", and is outside the
	 * 0-31 range.
	 *
	 * NOTE: Although this type could be described using a single
	 * byte (the maximum bit position possible is 31, after all),
	 * it is defined as an 'int' because 'char' types get printed
	 * out as chars, not integers, when used in '<<' debugging
	 * operators. If size becomes relevant, it can be changed to
	 * 'unsigned char'.
	 *
	 * FUTURE FIX: A better name for this type should be found!
	 **/
        typedef int BitPos;
        static const BitPos npos = 255;

	/**
	 * A type for use in methods that return a bit count.
	 *
	 */
        typedef unsigned BitCount;

        inline BitOverlay(Layout data) { this->bitsIs(data); }
        
        // Accessors
        inline Layout bits() const { return this->_bits; }
        
        // Interface

	/**
	 * Returns the bit position of the most-significant "on" bit.
	 *
	 * NOTE: Some chipsets have a 'find-first-set' (ffs)
	 * instruction that implements this. I've renamed it because
	 * I want to distinguish ffh and ffl.
	 *
	*/
        static BitPos FindFirstHigh(Layout val);
        inline BitPos findFirstHigh() const { return FindFirstHigh(this->_bits); }
        static BitPos FindFirstHighSlow(Layout val);

	/**
	 * Returns the bit position of the least-significant "on" bit.
	 *
	 */
        static BitPos FindFirstLow(Layout val);
        inline BitPos findFirstLow() const { return FindFirstLow(this->_bits); }

	/**
	 * Population Count: Returns the number of "on" bits.
	 *
	 */
	static inline int _bx(Layout x)             { return (x - ((x>>1)&0x77777777) - ((x>>2)&0x33333333) - ((x>>3)&0x11111111)); }
        static inline BitCount PopCount(Layout val) { return (((_bx(val)+(_bx(val)>>4)) & 0x0F0F0F0F) % 255); }
        inline BitCount popc() const                { return PopCount(this->_bits); }
        static BitCount PopCountSlow(Layout val);

	/**
	 * String representation of 'val' as a binary number.
	 *
	 * Currently always returns a string of length 32.  Future
	 * extensions may add an optional arg specifying when we
	 * can truncate (could use FFS to establish minimum length).
	 *
	 */
        static std::string Str(Layout val);
        inline std::string str() const { return Str(this->_bits); }

	/**
	 * Returns number of "regions" of contiguous "on" pixels.
	 *
	 * \param val The sequence of bits to analyze
	 * \param largest If non-null, it is assigned the length of the largest region.
	 * \param all If non-null, a pre-allocated array of size 16+, into which the sizes of each region are placed (may also encode start-bit-position).
	 *
	 */
        static BitCount Regions(Layout val, int* largest = NULL, int* all = NULL);
        inline BitCount regions(int* largest = NULL, int* all = NULL) const { return Regions(this->_bits, largest, all); }

	/**
	 * The Unit-testing method
	 *
	 */
        static void Test(int argc = 0, const char* argv[] = NULL);
        
      protected:
        // Accessors
        inline void bitsIs(const Layout & bits) { this->_bits = bits; }
        
        // Methods 
        
      private: 
        // Accessors
        inline Layout &  bitsRef() { return this->_bits; }
        
        // Methods 
        
        // State
        Layout    _bits;
    };
}

# endif // BitOverlay_h

