# ifndef CpuTimer_h
# define CpuTimer_h CpuTimer_h

# include "Root.h"  // parent

# include <sys/time.h>
# include <sys/resource.h>
# include <iomanip>  // setw

namespace Conjecture {

    // FUTURE FIX: Are these macros necessary?
# define TIME_USAGE(usage) \
	struct rusage usage; \
	if ( getrusage(RUSAGE_SELF,&usage) != 0 ) { cerr << "ERROR: getrusage failed!" << endl; exit(1); }
    
// end_usage   must be a 'struct rusage' variable
// start_usage must be a 'struct rusage' variable
// time must be a 'unsigned' variable
# define TIME_USAGE_DIFF(end_usage, start_usage, time) \
		time = \
		    ((now.ru_utime.tv_sec - start.ru_utime.tv_sec) * 1000000) + \
		    (now.ru_utime.tv_usec - start.ru_utime.tv_usec) + \
		    ((now.ru_stime.tv_sec - start.ru_stime.tv_sec) * 1000000) + \
		    (now.ru_stime.tv_usec - start.ru_stime.tv_usec); 
    
    // iters must be an unsigned number 
# define TIME_START_ITERATION(iters) { \
		 TIME_USAGE(start); \
		 for ( unsigned __i = 0; __i < iters; ++__i )
 
# define TIME_END_ITERATION(tdiff) \
	       TIME_USAGE(now); \
	       TIME_USAGE_DIFF(now, start, tdiff); \
	    }

    
    /********************************************************************
     * \class CpuTimer
     *
     * \brief microsecond-level timing support
     *
     * Provides a microsecond level CPU usage timer, for user
     * time, system time, or a combination of both. See also the
     * WallTimer class.
     *
     * Standard usage is:
     *    using namespace Conjecture;
     *    CpuTimer timer;  // turned on when created.
     *    ... 
     *    cout << "Total microsecond used: " << timer.millitime() << endl;
     *
     * Features are:
     *    timer.stop();   // stop timing so subsequent timing requests all
     *                    // apply to the same time interval
     *    timer.start();  // restart the timer from now.
     *    timer.print();  // prints out a full summary
     *    timer.XtimeY(); // Returns time interval in request units X (seconds
     *                    // if X is empty, otherwise X = 'milli' or 'micro').
     *                    // Only returns the time for request cpu section Y
     *                    // (both if Y is empty, otherwise Y = 'System' or 'User').
     *
     **********************************************************************/

    class CpuTimer : public Root {
      public:
	// Types

        // Constructors/Destructors
        CpuTimer();

        // Accessors
	inline const bool &          running() const { return this->_running; }
	inline const struct rusage & reading() const { return this->_reading; }
	inline const struct rusage & start()   const { return this->_start; }
        
        // Input/Output
        
        // Interface

	/**
	 * Start the timer (remember current time)
	 *
	 */
	void start();

	/**
	 * Stop the timer (remember current time)
	 *
	 */
	void stop();

	/**
	 * Returns the number of seconds between start and stop time.
	 *
	 */
        inline unsigned time()            const {return  read() / 1000000; }

	/**
	 * Returns the number of user seconds between start and stop time.
	 *
	 */
        inline unsigned timeUser()        const { return readUser() / 1000000; }

	/**
	 * Returns the number of system seconds between start and stop time.
	 *
	 */
        inline unsigned timeSystem()      const { return readSystem() / 1000000; }

	/**
	 * Returns the number of milliseconds between start and stop time.
	 *
	 */
        inline unsigned millitime()       const { return read() / 1000; }

	/**
	 * Returns the number of user milliseconds between start and stop time.
	 *
	 */
        inline unsigned millitimeUser()   const { return readUser() / 1000; }

	/**
	 * Returns the number of system milliseconds between start and stop time.
	 *
	 */
        inline unsigned millitimeSystem() const { return readSystem() / 1000; }

	/**
	 * Returns the number of microseconds between start and stop time.
	 *
	 */
        inline unsigned microtime()       const { return read() ; }

	/**
	 * Returns the number of user microseconds between start and stop time.
	 *
	 */
        inline unsigned microtimeUser()   const { return readUser() ; }

	/**
	 * Returns the number of system microseconds between start and stop time.
	 *
	 */
        inline unsigned microtimeSystem() const { return readSystem() ; }

	/**
	 * Print a description of the time results
	 *
	 */
	void print(ostream & os = cout) const;

	/**
	 * Print a one-line description of the time results
	 *
	 */
	void printSummary(ostream & os = cout) const;
	void printMilliSummary(ostream & os = cout) const;
	void printMicroSummary(ostream & os = cout) const;

	static void test(int argc = 0, const char* argv[] = NULL);
    
      protected:
        // Accessors
	inline void            runningIs(const bool & running) { this->_running = running; }
	inline void            readingIs(const struct rusage & reading) { this->_reading = reading; }
	inline void            startIs(const struct rusage & start) { this->_start = start; }
        
        // Methods 
        
      private: 
        // Accessors
	inline bool &          runningRef() { return this->_running; }
	inline struct rusage & readingRef() { return this->_reading; }
	inline struct rusage & startRef() { return this->_start; }
        
        // Methods 

	inline void getInfo( struct rusage & info ) const {
	    if ( getrusage(RUSAGE_SELF,&info) != 0 ) {
		cerr << "ERROR: getrusage failed!" << endl;
		// Throw an exception instead of crashing!
		exit(1);
	    }
	}

	unsigned long read() const;
	unsigned long readUser() const;
	unsigned long readSystem() const;
        
        // State
	bool            _running;
	struct rusage   _reading;
	struct rusage   _start;
    };
}

# endif // CpuTimer_h

